# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest
from datetime import datetime

from libcloud.test import MockHttp, LibcloudTestCase
from libcloud.utils.py3 import httplib
from libcloud.compute.base import NodeSize
from libcloud.test.secrets import GRIDSCALE_PARAMS
from libcloud.utils.iso8601 import UTC
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.compute.drivers.gridscale import GridscaleNodeDriver


class Gridscale_Tests(LibcloudTestCase):
    def setUp(self):
        GridscaleNodeDriver.connectionCls.conn_class = GridscaleMockHttp
        GridscaleMockHttp.type = None
        self.driver = GridscaleNodeDriver(*GRIDSCALE_PARAMS)

    def test_create_node_success(self):
        image = self.driver.list_images()[0]
        size = NodeSize(
            id=0,
            name="test",
            bandwidth=0,
            price=0,
            ram=10240,
            driver=self.driver,
            disk=50,
            extra={"cores": 2},
        )
        location = self.driver.list_locations()[0]
        sshkey = ["b1682d3a-1869-4bdc-8ffe-e74a261d300c"]
        GridscaleMockHttp.type = "POST"
        node = self.driver.create_node(
            name="test",
            size=size,
            image=image,
            location=location,
            ex_ssh_key_ids=sshkey,
        )

        self.assertEqual(node.name, "test.test")
        self.assertEqual(node.public_ips, ["185.102.95.236", "2a06:2380:0:1::211"])

    def test_create_image_success(self):
        node = self.driver.list_nodes()[0]
        GridscaleMockHttp.type = "POST"
        image = self.driver.create_image(node, "test.test")
        self.assertTrue(image)

    def test_list_nodes_success(self):
        nodes = self.driver.list_nodes()
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].name, "test.test")
        self.assertEqual(nodes[0].public_ips, ["185.102.95.236", "2a06:2380:0:1::211"])
        self.assertEqual(nodes[0].extra["cores"], 2)
        self.assertEqual(nodes[0].extra["memory"], 10240)

    def test_list_locations_success(self):
        locations = self.driver.list_locations()
        self.assertTrue(len(locations) >= 1)

    def test_list_volumes(self):
        volumes = self.driver.list_volumes()
        self.assertEqual(len(volumes), 1)
        volume = volumes[0]
        self.assertEqual(volume.id, "e66bb753-4a03-4ee2-a069-a601f393c9ee")
        self.assertEqual(volume.name, "linux")
        self.assertEqual(volume.size, 50)
        self.assertEqual(volume.driver, self.driver)

    def test_list_images_success(self):
        images = self.driver.list_images()
        self.assertTrue(len(images) >= 1)

        image = images[0]
        self.assertTrue(image.id is not None)
        self.assertTrue(image.name is not None)

    def test_list_key_pairs(self):
        keys = self.driver.list_key_pairs()
        self.assertEqual(len(keys), 2)
        self.assertEqual(keys[0].name, "karl")
        self.assertEqual(keys[0].public_key, "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQC")

    def test_list_volume_snapshots(self):
        volume = self.driver.list_volumes()[0]
        snapshots = self.driver.list_volume_snapshots(volume)

        self.assertEqual(len(snapshots), 1)
        self.assertEqual(snapshots[0].id, "d755de62-4d75-4d61-addd-a5c9743a5deb")

    def test_list_volumes_empty(self):
        GridscaleMockHttp.type = "EMPTY"
        volumes = self.driver.list_volumes()
        self.assertEqual(len(volumes), 0)

    def test_ex_list_networks(self):
        networks = self.driver.ex_list_networks()[0]
        self.assertEqual(networks.id, "1196529b-a8de-417f")

    def test_ex_list_ips(self):
        ip = self.driver.ex_list_ips()[0]
        self.assertEqual(ip.id, "56b8d161-325b-4fd4")

    def test_ex_destroy_ip(self):
        ip = self.driver.ex_list_ips()[0]

        GridscaleMockHttp.type = "DELETE"
        self.assertTrue(self.driver.ex_destroy_ip(ip))

    def test_ex_destroy_network(self):
        network = self.driver.ex_list_networks()[0]

        GridscaleMockHttp.type = "DELETE"
        self.assertTrue(self.driver.ex_destroy_network(network))

    def test_destroy_node_success(self):
        # Regular destroy
        node = self.driver.list_nodes()[0]
        GridscaleMockHttp.type = "DELETE"

        res = self.driver.destroy_node(node)
        self.assertTrue(res)

        # Destroy associated resources
        res = self.driver.destroy_node(node, ex_destroy_associated_resources=True)
        self.assertTrue(res)

    def test_destroy_volume(self):
        volume = self.driver.list_volumes()[0]
        GridscaleMockHttp.type = "DELETE"
        res = self.driver.destroy_volume(volume)
        self.assertTrue(res)

    def test_destroy_volume_snapshot(self):
        volume = self.driver.list_volumes()[0]
        snapshot = self.driver.list_volume_snapshots(volume)[0]
        GridscaleMockHttp.type = "DELETE"
        res = self.driver.destroy_volume_snapshot(snapshot)
        self.assertTrue(res)

    def test_get_image_success(self):
        image = self.driver.get_image("12345")
        self.assertEqual(image.id, "12345")

    def test_list_nodes_fills_created_datetime(self):
        nodes = self.driver.list_nodes()
        self.assertEqual(nodes[0].created_at, datetime(2019, 6, 7, 12, 56, 44, tzinfo=UTC))

    def test_ex_list_volumes_for_node(self):
        node = self.driver.list_nodes()[0]
        volumes = self.driver.ex_list_volumes_for_node(node=node)
        self.assertEqual(len(volumes), 1)
        self.assertEqual(volumes[0].size, 50)

    def test_ex_list_ips_for_node(self):
        node = self.driver.list_nodes()[0]
        ips = self.driver.ex_list_ips_for_node(node=node)
        self.assertEqual(len(ips), 1)
        self.assertEqual(ips[0].ip_address, "185.102.95.236")

    def test_ex_rename_node(self):
        node = self.driver.list_nodes()[0]
        self.assertTrue(self.driver.ex_rename_node(node, name="new-name"))

    def test_ex_rename_volume(self):
        volume = self.driver.list_volumes()[0]
        self.assertTrue(self.driver.ex_rename_volume(volume, name="new-name"))

    def test_ex_network(self):
        network = self.driver.ex_list_networks()[0]
        self.assertTrue(self.driver.ex_rename_network(network, name="new-name"))


class GridscaleMockHttp(MockHttp):
    fixtures = ComputeFileFixtures("gridscale")

    def _objects_servers(self, method, url, body, headers):
        body = self.fixtures.load("list_nodes.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_locations(self, method, url, body, headers):
        body = self.fixtures.load("list_locations.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_storages(self, method, url, body, headers):
        body = self.fixtures.load("list_volumes.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_storages_DELETE(self, method, url, body, headers):
        # test_destroy_node_success
        body = self.fixtures.load("list_volumes.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_ips_DELETE(self, method, url, body, headers):
        # test_destroy_node_success
        body = self.fixtures.load("ex_list_ips.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_ips_56b8d161_325b_4fd4_DELETE(self, method, url, body, headers):
        # test_destroy_node_success
        return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])

    def _objects_storages_EMPTY(self, method, url, body, headers):
        body = self.fixtures.load("list_volumes_empty.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_templates(self, method, url, body, headers):
        body = self.fixtures.load("list_images.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_sshkeys(self, method, url, body, headers):
        body = self.fixtures.load("list_key_pairs.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_storages_e66bb753_4a03_4ee2_a069_a601f393c9ee_snapshots(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("list_volume_snapshots.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_servers_1479405e_d46c_47a2_91e8_eb43951c899f(self, method, url, body, headers):
        # test_ex_rename_node
        if method == "PATCH":
            return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])
        else:  # pragma: no cover
            raise ValueError("Invalid method")

    def _objects_storages_e66bb753_4a03_4ee2_a069_a601f393c9ee(self, method, url, body, headers):
        # test_ex_rename_node
        if method == "PATCH":
            return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])
        else:  # pragma: no cover
            raise ValueError("Invalid method")

    def _objects_networks_1196529b_a8de_417f(self, method, url, body, headers):
        # test_ex_rename_network
        if method == "PATCH":
            return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])
        else:  # pragma: no cover
            raise ValueError("Invalid method")

    def _objects_servers_POST(self, method, url, body, headers):
        # create_node
        if method == "POST":
            body = self.fixtures.load("create_node.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:  # pragma: no cover
            raise ValueError("Invalid method")

    def _requests_x123xx1x_123x_1x12_123x_123xxx123x1x_POST(self, method, url, body, headers):
        body = self.fixtures.load("create_node_response_dict.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_servers_690de890_13c0_4e76_8a01_e10ba8786e53_POST(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("create_node_dict.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    # create_volume
    def _objects_storages_POST(self, method, url, body, headers):
        body = self.fixtures.load("create_volume.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_storages_690de890_13c0_4e76_8a01_e10ba8786e53_POST(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("create_volume_response_dict.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_ips_POST(self, method, url, body, headers):
        # ex_create_ip
        if method == "POST":
            body = self.fixtures.load("create_ip.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:  # pragma: no cover
            raise ValueError("Invalid method")

    def _objects_ips_690de890_13c0_4e76_8a01_e10ba8786e53_POST(self, method, url, body, headers):
        body = self.fixtures.load("create_ip_response_dict.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_servers_1479405e_d46c_47a2_91e8_eb43951c899f_storages_POST(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("volume_to_node.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_servers_1479405e_d46c_47a2_91e8_eb43951c899f_ips_POST(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("ips_to_node.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_networks_POST(self, method, url, body, headers):
        body = self.fixtures.load("create_network.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_networks_1196529b_a8de_417f_DELETE(self, method, url, body, headers):
        # test_ex_destroy_network
        if method == "DELETE":
            return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])
        else:  # pragma: no cover
            raise ValueError("Invalid method")

    def _objects_servers_1479405e_d46c_47a2_91e8_eb43951c899f_networks_POST(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("network_to_node.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_servers_1479405e_d46c_47a2_91e8_eb43951c899f_power_POST(
        self, method, url, body, headers
    ):
        if method == "PATCH":
            body = self.fixtures.load("ex_start_node.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_servers_1479405e_d46c_47a2_91e8_eb43951c899f_POST(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("create_node_dict.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_networks(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("ex_list_networks.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:  # pragma: no cover
            raise ValueError("Invalid method")

    def _objects_ips(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("ex_list_ips.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:  # pragma: no cover
            raise ValueError("Invalid method")

    def _objects_servers_1479405e_d46c_47a2_91e8_eb43951c899f_DELETE(
        self, method, url, body, headers
    ):
        if method == "DELETE":
            return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])
        else:  # pragma: no cover
            raise ValueError("Invalid method")

    def _objects_storages_e66bb753_4a03_4ee2_a069_a601f393c9ee_DELETE(
        self, method, url, body, headers
    ):
        if method == "DELETE":
            return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])
        else:  # pragma: no cover
            raise ValueError("Invalid method")

    def _objects_storages_e66bb753_4a03_4ee2_a069_a601f393c9ee_snapshots_d755de62_4d75_4d61_addd_a5c9743a5deb_DELETE(
        self, method, url, body, headers
    ):
        if method == "DELETE":
            return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])
        else:  # pragma: no cover
            raise ValueError("Invalid method")

    def _objects_templates_POST(self, method, url, body, headers):
        # create_image
        if method == "POST":
            body = self.fixtures.load("create_image.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:  # pragma: no cover
            raise ValueError("Invalid method")

    def _objects_storages_e66bb753_4a03_4ee2_a069_a601f393c9ee_snapshots_POST(
        self, method, url, body, headers
    ):
        # create_image
        if method == "POST":
            body = self.fixtures.load("create_image.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:  # pragma: no cover
            raise ValueError("Invalid method")

    def _objects_storages_e66bb753_4a03_4ee2_a069_a601f393c9ee_snapshots_690de890_13c0_4e76_8a01_e10ba8786e53_POST(
        self, method, url, body, headers
    ):
        # create_image
        body = self.fixtures.load("create_image_dict.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_templates_690de890_13c0_4e76_8a01_e10ba8786e53_POST(
        self, method, url, body, headers
    ):
        # create_image
        if method == "GET":
            body = self.fixtures.load("create_image_dict.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:  # pragma: no cover
            raise ValueError("Invalid method")

    def _objects_templates_12345(self, method, url, body, headers):
        # get_image
        if method == "GET":
            body = self.fixtures.load("get_image.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:  # pragma: no cover
            raise ValueError("Invalid method")


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more§
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import ssl
import sys
import socket
from unittest import mock
from unittest.mock import Mock, patch

import requests_mock
from requests.exceptions import ConnectTimeout

import libcloud.common.base
from libcloud.http import LibcloudConnection, SignedHTTPSAdapter, LibcloudBaseConnection
from libcloud.test import unittest, no_internet
from libcloud.utils.py3 import assertRaisesRegex
from libcloud.common.base import Response, Connection, CertificateConnection
from libcloud.utils.retry import RETRY_EXCEPTIONS, Retry, RetryForeverOnRateLimitError
from libcloud.common.exceptions import RateLimitReachedError


class BaseConnectionClassTestCase(unittest.TestCase):
    def setUp(self):
        self.orig_http_proxy = os.environ.pop("http_proxy", None)
        self.orig_https_proxy = os.environ.pop("https_proxy", None)

    def tearDown(self):
        if self.orig_http_proxy:
            os.environ["http_proxy"] = self.orig_http_proxy
        elif "http_proxy" in os.environ:
            del os.environ["http_proxy"]

        if self.orig_https_proxy:
            os.environ["https_proxy"] = self.orig_https_proxy
        elif "https_proxy" in os.environ:
            del os.environ["https_proxy"]

        libcloud.common.base.ALLOW_PATH_DOUBLE_SLASHES = False

    def test_parse_proxy_url(self):
        conn = LibcloudBaseConnection()

        proxy_url = "http://127.0.0.1:3128"
        result = conn._parse_proxy_url(proxy_url=proxy_url)
        self.assertEqual(result[0], "http")
        self.assertEqual(result[1], "127.0.0.1")
        self.assertEqual(result[2], 3128)
        self.assertIsNone(result[3])
        self.assertIsNone(result[4])

        proxy_url = "https://127.0.0.2:3129"
        result = conn._parse_proxy_url(proxy_url=proxy_url)
        self.assertEqual(result[0], "https")
        self.assertEqual(result[1], "127.0.0.2")
        self.assertEqual(result[2], 3129)
        self.assertIsNone(result[3])
        self.assertIsNone(result[4])

        proxy_url = "http://user1:pass1@127.0.0.1:3128"
        result = conn._parse_proxy_url(proxy_url=proxy_url)
        self.assertEqual(result[0], "http")
        self.assertEqual(result[1], "127.0.0.1")
        self.assertEqual(result[2], 3128)
        self.assertEqual(result[3], "user1")
        self.assertEqual(result[4], "pass1")

        proxy_url = "https://user1:pass1@127.0.0.2:3129"
        result = conn._parse_proxy_url(proxy_url=proxy_url)
        self.assertEqual(result[0], "https")
        self.assertEqual(result[1], "127.0.0.2")
        self.assertEqual(result[2], 3129)
        self.assertEqual(result[3], "user1")
        self.assertEqual(result[4], "pass1")

        proxy_url = "http://127.0.0.1"
        expected_msg = "proxy_url must be in the following format"
        assertRaisesRegex(
            self, ValueError, expected_msg, conn._parse_proxy_url, proxy_url=proxy_url
        )

        proxy_url = "http://@127.0.0.1:3128"
        expected_msg = "URL is in an invalid format"
        assertRaisesRegex(
            self, ValueError, expected_msg, conn._parse_proxy_url, proxy_url=proxy_url
        )

        proxy_url = "http://user@127.0.0.1:3128"
        expected_msg = "URL is in an invalid format"
        assertRaisesRegex(
            self, ValueError, expected_msg, conn._parse_proxy_url, proxy_url=proxy_url
        )

    def test_constructor(self):
        proxy_url = "http://127.0.0.2:3128"
        os.environ["http_proxy"] = proxy_url
        conn = LibcloudConnection(host="localhost", port=80)
        self.assertEqual(conn.proxy_scheme, "http")
        self.assertEqual(conn.proxy_host, "127.0.0.2")
        self.assertEqual(conn.proxy_port, 3128)
        self.assertEqual(
            conn.session.proxies,
            {"http": "http://127.0.0.2:3128", "https": "http://127.0.0.2:3128"},
        )

        _ = os.environ.pop("http_proxy", None)
        conn = LibcloudConnection(host="localhost", port=80)
        self.assertIsNone(conn.proxy_scheme)
        self.assertIsNone(conn.proxy_host)
        self.assertIsNone(conn.proxy_port)

        proxy_url = "http://127.0.0.3:3128"
        conn.set_http_proxy(proxy_url=proxy_url)
        self.assertEqual(conn.proxy_scheme, "http")
        self.assertEqual(conn.proxy_host, "127.0.0.3")
        self.assertEqual(conn.proxy_port, 3128)
        self.assertEqual(
            conn.session.proxies,
            {"http": "http://127.0.0.3:3128", "https": "http://127.0.0.3:3128"},
        )

        proxy_url = "http://127.0.0.4:3128"
        conn = LibcloudConnection(host="localhost", port=80, proxy_url=proxy_url)
        self.assertEqual(conn.proxy_scheme, "http")
        self.assertEqual(conn.proxy_host, "127.0.0.4")
        self.assertEqual(conn.proxy_port, 3128)
        self.assertEqual(
            conn.session.proxies,
            {"http": "http://127.0.0.4:3128", "https": "http://127.0.0.4:3128"},
        )

        os.environ["http_proxy"] = proxy_url
        proxy_url = "http://127.0.0.5:3128"
        conn = LibcloudConnection(host="localhost", port=80, proxy_url=proxy_url)
        self.assertEqual(conn.proxy_scheme, "http")
        self.assertEqual(conn.proxy_host, "127.0.0.5")
        self.assertEqual(conn.proxy_port, 3128)
        self.assertEqual(
            conn.session.proxies,
            {"http": "http://127.0.0.5:3128", "https": "http://127.0.0.5:3128"},
        )

        os.environ["http_proxy"] = proxy_url
        proxy_url = "https://127.0.0.6:3129"
        conn = LibcloudConnection(host="localhost", port=80, proxy_url=proxy_url)
        self.assertEqual(conn.proxy_scheme, "https")
        self.assertEqual(conn.proxy_host, "127.0.0.6")
        self.assertEqual(conn.proxy_port, 3129)
        self.assertEqual(
            conn.session.proxies,
            {"http": "https://127.0.0.6:3129", "https": "https://127.0.0.6:3129"},
        )

    def test_connection_to_unusual_port(self):
        conn = LibcloudConnection(host="localhost", port=8080)
        self.assertIsNone(conn.proxy_scheme)
        self.assertIsNone(conn.proxy_host)
        self.assertIsNone(conn.proxy_port)
        self.assertEqual(conn.host, "http://localhost:8080")

        conn = LibcloudConnection(host="localhost", port=80)
        self.assertEqual(conn.host, "http://localhost")

    def test_connection_session_timeout(self):
        """
        Test that the connection timeout attribute is set correctly
        """
        conn = LibcloudConnection(host="localhost", port=8080)
        self.assertEqual(conn.session.timeout, 60)

        conn = LibcloudConnection(host="localhost", port=8080, timeout=10)
        self.assertEqual(conn.session.timeout, 10)

    @unittest.skipIf(no_internet(), "Internet is not reachable")
    def test_connection_timeout_raised(self):
        """
        Test that the connection times out
        """
        conn = LibcloudConnection(host="localhost", port=8080, timeout=0.1)
        # use a not-routable address to test that the connection timeouts
        host = "http://10.255.255.1"
        with self.assertRaises(ConnectTimeout):
            conn.request("GET", host)

    def test_connection_url_merging(self):
        """
        Test that the connection class will parse URLs correctly
        """
        conn = Connection(url="http://test.com/")
        conn.connect()
        self.assertEqual(conn.connection.host, "http://test.com")
        with requests_mock.mock() as m:
            m.get("http://test.com/test", text="data")
            response = conn.request("/test")
        self.assertEqual(response.body, "data")

    def test_morph_action_hook(self):
        conn = Connection(url="http://test.com")

        conn.request_path = ""
        self.assertEqual(conn.morph_action_hook("/test"), "/test")
        self.assertEqual(conn.morph_action_hook("test"), "/test")

        conn.request_path = "/v1"
        self.assertEqual(conn.morph_action_hook("/test"), "/v1/test")
        self.assertEqual(conn.morph_action_hook("test"), "/v1/test")

        conn.request_path = "/v1"
        self.assertEqual(conn.morph_action_hook("/test"), "/v1/test")
        self.assertEqual(conn.morph_action_hook("test"), "/v1/test")

        conn.request_path = "v1"
        self.assertEqual(conn.morph_action_hook("/test"), "/v1/test")
        self.assertEqual(conn.morph_action_hook("test"), "/v1/test")

        conn.request_path = "v1/"
        self.assertEqual(conn.morph_action_hook("/test"), "/v1/test")
        self.assertEqual(conn.morph_action_hook("test"), "/v1/test")

        conn.request_path = "/a"
        self.assertEqual(conn.morph_action_hook("//b/c.txt"), "/a/b/c.txt")

        conn.request_path = "/b"
        self.assertEqual(conn.morph_action_hook("/foo//"), "/b/foo/")

        libcloud.common.base.ALLOW_PATH_DOUBLE_SLASHES = True

        conn.request_path = "/"
        self.assertEqual(conn.morph_action_hook("/"), "//")

        conn.request_path = ""
        self.assertEqual(conn.morph_action_hook("/"), "/")

        conn.request_path = "/a"
        self.assertEqual(conn.morph_action_hook("//b/c.txt"), "/a//b/c.txt")

        conn.request_path = "/b"
        self.assertEqual(conn.morph_action_hook("/foo//"), "/b/foo//")

    def test_connect_with_prefix(self):
        """
        Test that a connection with a base path (e.g. /v1/) will
        add the base path to requests
        """
        conn = Connection(url="http://test.com/")
        conn.connect()
        conn.request_path = "/v1"
        self.assertEqual(conn.connection.host, "http://test.com")
        with requests_mock.mock() as m:
            m.get("http://test.com/v1/test", text="data")
            response = conn.request("/test")
        self.assertEqual(response.body, "data")

    def test_secure_connection_unusual_port(self):
        """
        Test that the connection class will default to secure (https) even
        when the port is an unusual (non 443, 80) number
        """
        conn = Connection(secure=True, host="localhost", port=8081)
        conn.connect()
        self.assertEqual(conn.connection.host, "https://localhost:8081")

        conn2 = Connection(url="https://localhost:8081")
        conn2.connect()
        self.assertEqual(conn2.connection.host, "https://localhost:8081")

    def test_secure_by_default(self):
        """
        Test that the connection class will default to secure (https)
        """
        conn = Connection(host="localhost", port=8081)
        conn.connect()
        self.assertEqual(conn.connection.host, "https://localhost:8081")

    def test_implicit_port(self):
        """
        Test that the port is not included in the URL if the protocol implies
        the port, e.g. http implies 80
        """
        conn = Connection(secure=True, host="localhost", port=443)
        conn.connect()
        self.assertEqual(conn.connection.host, "https://localhost")

        conn2 = Connection(secure=False, host="localhost", port=80)
        conn2.connect()
        self.assertEqual(conn2.connection.host, "http://localhost")

    def test_insecure_connection_unusual_port(self):
        """
        Test that the connection will allow unusual ports and insecure
        schemes
        """
        conn = Connection(secure=False, host="localhost", port=8081)
        conn.connect()
        self.assertEqual(conn.connection.host, "http://localhost:8081")

        conn2 = Connection(url="http://localhost:8081")
        conn2.connect()
        self.assertEqual(conn2.connection.host, "http://localhost:8081")


class ConnectionClassTestCase(unittest.TestCase):
    def setUp(self):
        self.originalConnect = Connection.connect
        self.originalResponseCls = Connection.responseCls

        Connection.connect = Mock()
        Connection.responseCls = Mock()
        Connection.allow_insecure = True

    def tearDown(self):
        Connection.connect = self.originalConnect
        Connection.responseCls = Connection.responseCls
        Connection.allow_insecure = True

    def test_dont_allow_insecure(self):
        Connection.allow_insecure = True
        Connection(secure=False)

        Connection.allow_insecure = False

        expected_msg = r"Non https connections are not allowed \(use " r"secure=True\)"
        assertRaisesRegex(self, ValueError, expected_msg, Connection, secure=False)

    def test_cache_busting(self):
        params1 = {"foo1": "bar1", "foo2": "bar2"}
        params2 = [("foo1", "bar1"), ("foo2", "bar2")]

        con = Connection()
        con.connection = Mock()
        con.pre_connect_hook = Mock()
        con.pre_connect_hook.return_value = {}, {}
        con.cache_busting = False

        con.request(action="/path", params=params1)
        args, kwargs = con.pre_connect_hook.call_args
        self.assertFalse("cache-busting" in args[0])
        self.assertEqual(args[0], params1)

        con.request(action="/path", params=params2)
        args, kwargs = con.pre_connect_hook.call_args
        self.assertFalse("cache-busting" in args[0])
        self.assertEqual(args[0], params2)

        con.cache_busting = True

        con.request(action="/path", params=params1)
        args, kwargs = con.pre_connect_hook.call_args
        self.assertTrue("cache-busting" in args[0])

        con.request(action="/path", params=params2)
        args, kwargs = con.pre_connect_hook.call_args
        self.assertTrue("cache-busting" in args[0][len(params2)])

    def test_context_is_reset_after_request_has_finished(self):
        context = {"foo": "bar"}

        def responseCls(connection, response) -> mock.MagicMock:
            connection.called = True
            self.assertEqual(connection.context, context)
            return mock.MagicMock(spec=Response)

        con = Connection()
        con.called = False
        con.connection = Mock()
        con.responseCls = responseCls

        con.set_context(context)
        self.assertEqual(con.context, context)

        con.request("/")

        # Context should have been reset
        self.assertTrue(con.called)
        self.assertEqual(con.context, {})

        # Context should also be reset if a method inside request throws
        con = Connection(timeout=1, retry_delay=0.1)
        con.connection = Mock()

        con.set_context(context)
        self.assertEqual(con.context, context)
        con.connection.request = Mock(side_effect=ssl.SSLError())

        try:
            con.request("/")
        except ssl.SSLError:
            pass

        self.assertEqual(con.context, {})

        con.connection = Mock()
        con.set_context(context)
        self.assertEqual(con.context, context)

        con.responseCls = Mock(side_effect=ValueError())

        try:
            con.request("/")
        except ValueError:
            pass

        self.assertEqual(con.context, {})

    def _raise_socket_error(self):
        raise socket.gaierror("")

    @patch("libcloud.common.base.Connection.request")
    def test_retry_with_sleep(self, mock_connect):
        con = Connection()
        con.connection = Mock()

        mock_connect.side_effect = socket.gaierror("")
        retry_request = Retry(timeout=1, retry_delay=0.1, backoff=1)
        self.assertRaises(socket.gaierror, retry_request(con.request), action="/")

        self.assertGreater(mock_connect.call_count, 1, "Retry logic failed")

    @patch("libcloud.common.base.Connection.request")
    def test_retry_with_timeout(self, mock_connect):
        con = Connection()
        con.connection = Mock()

        mock_connect.side_effect = socket.gaierror("")
        retry_request = Retry(timeout=1, retry_delay=0.1, backoff=1)
        self.assertRaises(socket.gaierror, retry_request(con.request), action="/")

        self.assertGreater(mock_connect.call_count, 1, "Retry logic failed")

    @patch("libcloud.common.base.Connection.request")
    def test_retry_with_backoff(self, mock_connect):
        con = Connection()
        con.connection = Mock()

        mock_connect.side_effect = socket.gaierror("")
        retry_request = Retry(timeout=1, retry_delay=0.1, backoff=1)
        self.assertRaises(socket.gaierror, retry_request(con.request), action="/")
        self.assertGreater(mock_connect.call_count, 1, "Retry logic failed")

    @patch("libcloud.common.base.Connection.request")
    def test_retry_rate_limit_error_timeout(self, mock_connect):
        con = Connection()
        con.connection = Mock()

        mock_connect.__name__ = "mock_connect"
        headers = {"retry-after": 0.2}
        mock_connect.side_effect = RateLimitReachedError(headers=headers)
        retry_request = Retry(timeout=1, retry_delay=0.1, backoff=1)
        self.assertRaises(RateLimitReachedError, retry_request(con.request), action="/")
        self.assertGreater(mock_connect.call_count, 1, "Retry logic failed")

    @patch("libcloud.common.base.Connection.request")
    def test_retry_rate_limit_error_forever_with_old_retry_class(self, mock_connect):
        con = Connection()
        con.connection = Mock()

        self.retry_counter = 0

        def mock_connect_side_effect(*args, **kwargs):
            self.retry_counter += 1

            if self.retry_counter < 4:
                headers = {"retry-after": 0.1}
                raise RateLimitReachedError(headers=headers)

            return "success"

        mock_connect.__name__ = "mock_connect"
        mock_connect.side_effect = mock_connect_side_effect
        retry_request = RetryForeverOnRateLimitError(timeout=1, retry_delay=0.1, backoff=1)
        retry_request(con.request)(action="/")

        # We have waited longer the timeout but continue to retry
        result = retry_request(con.request)(action="/")
        self.assertEqual(result, "success")

        self.assertEqual(mock_connect.call_count, 5, "Retry logic failed")

    @patch("libcloud.common.base.Connection.request")
    def test_retry_should_not_retry_on_non_defined_exception(self, mock_connect):
        con = Connection()
        con.connection = Mock()

        self.retry_counter = 0

        mock_connect.__name__ = "mock_connect"
        mock_connect.side_effect = ValueError("should not retry this " "error")
        retry_request = Retry(timeout=5, retry_delay=0.1, backoff=1)

        self.assertRaisesRegex(
            ValueError,
            "should not retry this error",
            retry_request(con.request),
            action="/",
        )
        self.assertEqual(mock_connect.call_count, 1, "Retry logic failed")

    @patch("libcloud.common.base.Connection.request")
    def test_retry_rate_limit_error_success_on_second_attempt(self, mock_connect):
        con = Connection()
        con.connection = Mock()

        self.retry_counter = 0

        def mock_connect_side_effect(*args, **kwargs):
            self.retry_counter += 1

            if self.retry_counter < 2:
                headers = {"retry-after": 0.2}
                raise RateLimitReachedError(headers=headers)

            return "success"

        mock_connect.__name__ = "mock_connect"
        mock_connect.side_effect = mock_connect_side_effect
        retry_request = Retry(timeout=1, retry_delay=0.1, backoff=1)
        result = retry_request(con.request)(action="/")
        self.assertEqual(result, "success")

        self.assertEqual(mock_connect.call_count, 2, "Retry logic failed")

    @patch("libcloud.common.base.Connection.request")
    def test_retry_on_all_default_retry_exception_classes(self, mock_connect):
        con = Connection()
        con.connection = Mock()

        self.retry_counter = 0

        def mock_connect_side_effect(*args, **kwargs):
            self.retry_counter += 1

            if self.retry_counter < len(RETRY_EXCEPTIONS):
                raise RETRY_EXCEPTIONS[self.retry_counter]

            return "success"

        mock_connect.__name__ = "mock_connect"
        mock_connect.side_effect = mock_connect_side_effect
        retry_request = Retry(timeout=1, retry_delay=0.1, backoff=1)
        result = retry_request(con.request)(action="/")
        self.assertEqual(result, "success")

        self.assertEqual(mock_connect.call_count, len(RETRY_EXCEPTIONS), "Retry logic failed")

    def test_request_parses_errors(self):
        class ThrowingResponse(Response):
            def __init__(self, *_, **__):
                super().__init__(mock.MagicMock(), mock.MagicMock())

            def parse_body(self):
                return super().parse_body()

            def parse_error(self):
                raise RateLimitReachedError()

            def success(self):
                return False

        con = Connection()
        con.connection = Mock()
        con.responseCls = ThrowingResponse

        with self.assertRaises(RateLimitReachedError):
            con.request(action="/")

    def test_parse_errors_can_be_retried(self):
        class RetryableThrowingError(Response):
            parse_error_counter: int = 0
            success_counter: int = 0

            def __init__(self, *_, **__):
                super().__init__(mock.MagicMock(), mock.MagicMock())

            def parse_body(self):
                return super().parse_body()

            def parse_error(self):
                RetryableThrowingError.parse_error_counter += 1
                if RetryableThrowingError.parse_error_counter > 1:
                    return "success"
                else:
                    raise RateLimitReachedError()

            def success(self):
                RetryableThrowingError.success_counter += 1
                if RetryableThrowingError.success_counter > 1:
                    return True
                else:
                    return False

        con = Connection()
        con.connection = Mock()
        con.responseCls = RetryableThrowingError
        result = con.request(action="/", retry_failed=True)

        self.assertEqual(result.success(), True)


class CertificateConnectionClassTestCase(unittest.TestCase):
    def setUp(self):
        self.connection = CertificateConnection(cert_file="test.pem", url="https://test.com/test")
        self.connection.connect()

    def test_adapter_internals(self):
        adapter = self.connection.connection.session.adapters["https://"]
        self.assertTrue(isinstance(adapter, SignedHTTPSAdapter))
        self.assertEqual(adapter.cert_file, "test.pem")


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import copy
import hmac
import math
import hashlib
import os.path  # pylint: disable-msg=W0404
from io import BytesIO
from hashlib import sha1
from unittest import mock
from unittest.mock import Mock, PropertyMock

import libcloud.utils.files
from libcloud.test import MockHttp  # pylint: disable-msg=E0611
from libcloud.test import unittest, make_response, generate_random_data
from libcloud.utils.py3 import StringIO, b, httplib, urlquote
from libcloud.utils.files import exhaust_iterator
from libcloud.common.types import MalformedResponseError
from libcloud.storage.base import CHUNK_SIZE, Object, Container
from libcloud.storage.types import (
    ObjectDoesNotExistError,
    ObjectHashMismatchError,
    ContainerIsNotEmptyError,
    InvalidContainerNameError,
    ContainerDoesNotExistError,
    ContainerAlreadyExistsError,
)
from libcloud.test.storage.base import BaseRangeDownloadMockHttp
from libcloud.test.file_fixtures import StorageFileFixtures  # pylint: disable-msg=E0611
from libcloud.storage.drivers.cloudfiles import CloudFilesStorageDriver


class CloudFilesTests(unittest.TestCase):
    driver_klass = CloudFilesStorageDriver
    driver_args = ("dummy", "dummy")
    driver_kwargs = {}
    region = "ord"

    def setUp(self):
        self.driver_klass.connectionCls.conn_class = CloudFilesMockHttp
        CloudFilesMockHttp.type = None

        driver_kwargs = self.driver_kwargs.copy()
        driver_kwargs["region"] = self.region
        self.driver = self.driver_klass(*self.driver_args, **driver_kwargs)

        # normally authentication happens lazily, but we force it here
        self.driver.connection._populate_hosts_and_request_paths()
        self._remove_test_file()

    def tearDown(self):
        self._remove_test_file()

    def test_invalid_ex_force_service_region(self):
        driver = CloudFilesStorageDriver("driver", "dummy", ex_force_service_region="invalid")

        try:
            driver.list_containers()
        except Exception as e:
            self.assertEqual(e.value, "Could not find specified endpoint")
        else:
            self.fail("Exception was not thrown")

    def test_ex_force_service_region(self):
        driver = CloudFilesStorageDriver("driver", "dummy", ex_force_service_region="ORD")
        driver.list_containers()

    def test_force_auth_token_kwargs(self):
        base_url = "https://cdn2.clouddrive.com/v1/MossoCloudFS"
        kwargs = {
            "ex_force_auth_token": "some-auth-token",
            "ex_force_base_url": base_url,
        }
        driver = CloudFilesStorageDriver("driver", "dummy", **kwargs)
        driver.list_containers()

        self.assertEqual(kwargs["ex_force_auth_token"], driver.connection.auth_token)
        self.assertEqual("cdn2.clouddrive.com", driver.connection.host)
        self.assertEqual("/v1/MossoCloudFS", driver.connection.request_path)

    def test_force_auth_url_kwargs(self):
        kwargs = {
            "ex_force_auth_version": "2.0",
            "ex_force_auth_url": "https://identity.api.rackspace.com",
        }
        driver = CloudFilesStorageDriver("driver", "dummy", **kwargs)

        self.assertEqual(kwargs["ex_force_auth_url"], driver.connection._ex_force_auth_url)
        self.assertEqual(kwargs["ex_force_auth_version"], driver.connection._auth_version)

    def test_invalid_json_throws_exception(self):
        CloudFilesMockHttp.type = "MALFORMED_JSON"
        try:
            self.driver.list_containers()
        except MalformedResponseError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_service_catalog(self):
        url = "https://storage4.%s1.clouddrive.com/v1/MossoCloudFS" % (self.region)
        self.assertEqual(url, self.driver.connection.get_endpoint())

        self.driver.connection.cdn_request = True
        self.assertEqual(
            "https://cdn.clouddrive.com/v1/MossoCloudFS",
            self.driver.connection.get_endpoint(),
        )
        self.driver.connection.cdn_request = False

    def test_get_endpoint_internalurl(self):
        self.driver.connection.use_internal_url = True
        url = (
            "https://snet-storage101.%s1.clouddrive.com/v1/MossoCloudFS_11111-111111111-1111111111-1111111"
            % (self.region)
        )
        self.assertEqual(url, self.driver.connection.get_endpoint())

    def test_list_containers(self):
        CloudFilesMockHttp.type = "EMPTY"
        containers = self.driver.list_containers()
        self.assertEqual(len(containers), 0)

        CloudFilesMockHttp.type = None
        containers = self.driver.list_containers()
        self.assertEqual(len(containers), 3)

        container = [c for c in containers if c.name == "container2"][0]
        self.assertEqual(container.extra["object_count"], 120)
        self.assertEqual(container.extra["size"], 340084450)

    def test_list_container_objects(self):
        CloudFilesMockHttp.type = "EMPTY"
        container = Container(name="test_container", extra={}, driver=self.driver)
        objects = self.driver.list_container_objects(container=container)
        self.assertEqual(len(objects), 0)

        CloudFilesMockHttp.type = None
        objects = self.driver.list_container_objects(container=container)
        self.assertEqual(len(objects), 4)

        obj = [o for o in objects if o.name == "foo test 1"][0]
        self.assertEqual(obj.hash, "16265549b5bda64ecdaa5156de4c97cc")
        self.assertEqual(obj.size, 1160520)
        self.assertEqual(obj.container.name, "test_container")

    def test_list_container_object_name_encoding(self):
        CloudFilesMockHttp.type = "EMPTY"
        container = Container(name="test container 1", extra={}, driver=self.driver)
        objects = self.driver.list_container_objects(container=container)
        self.assertEqual(len(objects), 0)

    def test_list_container_objects_with_prefix(self):
        CloudFilesMockHttp.type = "EMPTY"
        container = Container(name="test_container", extra={}, driver=self.driver)
        objects = self.driver.list_container_objects(container=container, prefix="test_prefix1")
        self.assertEqual(len(objects), 0)

        CloudFilesMockHttp.type = None
        objects = self.driver.list_container_objects(container=container, prefix="test_prefix2")
        self.assertEqual(len(objects), 4)

        obj = [o for o in objects if o.name == "foo test 1"][0]
        self.assertEqual(obj.hash, "16265549b5bda64ecdaa5156de4c97cc")
        self.assertEqual(obj.size, 1160520)
        self.assertEqual(obj.container.name, "test_container")

    def test_list_container_objects_iterator(self):
        CloudFilesMockHttp.type = "ITERATOR"
        container = Container(name="test_container", extra={}, driver=self.driver)
        objects = self.driver.list_container_objects(container=container)
        self.assertEqual(len(objects), 5)

        obj = [o for o in objects if o.name == "foo-test-1"][0]
        self.assertEqual(obj.hash, "16265549b5bda64ecdaa5156de4c97cc")
        self.assertEqual(obj.size, 1160520)
        self.assertEqual(obj.container.name, "test_container")

    def test_get_container(self):
        container = self.driver.get_container(container_name="test_container")
        self.assertEqual(container.name, "test_container")
        self.assertEqual(container.extra["object_count"], 800)
        self.assertEqual(container.extra["size"], 1234568)

    def test_get_container_not_found(self):
        try:
            self.driver.get_container(container_name="not_found")
        except ContainerDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_get_object_success(self):
        obj = self.driver.get_object(container_name="test_container", object_name="test_object")
        self.assertEqual(obj.container.name, "test_container")
        self.assertEqual(obj.size, 555)
        self.assertEqual(obj.hash, "6b21c4a111ac178feacf9ec9d0c71f17")
        self.assertEqual(obj.extra["content_type"], "application/zip")
        self.assertEqual(obj.extra["last_modified"], "Tue, 25 Jan 2011 22:01:49 GMT")
        self.assertEqual(obj.meta_data["foo-bar"], "test 1")
        self.assertEqual(obj.meta_data["bar-foo"], "test 2")

    def test_get_object_object_name_encoding(self):
        obj = self.driver.get_object(container_name="test_container", object_name="~/test_object/")
        self.assertEqual(obj.name, "~/test_object/")

    def test_get_object_not_found(self):
        try:
            self.driver.get_object(container_name="test_container", object_name="not_found")
        except ObjectDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_create_container_success(self):
        container = self.driver.create_container(container_name="test_create_container")
        self.assertTrue(isinstance(container, Container))
        self.assertEqual(container.name, "test_create_container")
        self.assertEqual(container.extra["object_count"], 0)

    def test_create_container_already_exists(self):
        CloudFilesMockHttp.type = "ALREADY_EXISTS"

        try:
            self.driver.create_container(container_name="test_create_container")
        except ContainerAlreadyExistsError:
            pass
        else:
            self.fail("Container already exists but an exception was not thrown")

    def test_create_container_invalid_name_too_long(self):
        name = "".join(["x" for x in range(0, 257)])
        try:
            self.driver.create_container(container_name=name)
        except InvalidContainerNameError:
            pass
        else:
            self.fail(
                "Invalid name was provided (name is too long)" ", but exception was not thrown"
            )

    def test_create_container_invalid_name_slashes_in_name(self):
        try:
            self.driver.create_container(container_name="test/slashes/")
        except InvalidContainerNameError:
            pass
        else:
            self.fail(
                "Invalid name was provided (name contains slashes)" ", but exception was not thrown"
            )

    def test_delete_container_success(self):
        container = Container(name="foo_bar_container", extra={}, driver=self)
        result = self.driver.delete_container(container=container)
        self.assertTrue(result)

    def test_delete_container_not_found(self):
        CloudFilesMockHttp.type = "NOT_FOUND"
        container = Container(name="foo_bar_container", extra={}, driver=self)
        try:
            self.driver.delete_container(container=container)
        except ContainerDoesNotExistError:
            pass
        else:
            self.fail("Container does not exist but an exception was not thrown")

    def test_delete_container_not_empty(self):
        CloudFilesMockHttp.type = "NOT_EMPTY"
        container = Container(name="foo_bar_container", extra={}, driver=self)
        try:
            self.driver.delete_container(container=container)
        except ContainerIsNotEmptyError:
            pass
        else:
            self.fail("Container is not empty but an exception was not thrown")

    def test_download_object_success(self):
        container = Container(name="foo_bar_container", extra={}, driver=self)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=CloudFilesStorageDriver,
        )
        destination_path = os.path.abspath(__file__) + ".temp"
        result = self.driver.download_object(
            obj=obj,
            destination_path=destination_path,
            overwrite_existing=False,
            delete_on_failure=True,
        )
        self.assertTrue(result)

    def test_download_object_invalid_file_size(self):
        CloudFilesMockHttp.type = "INVALID_SIZE"
        container = Container(name="foo_bar_container", extra={}, driver=self)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=CloudFilesStorageDriver,
        )
        destination_path = os.path.abspath(__file__) + ".temp"
        result = self.driver.download_object(
            obj=obj,
            destination_path=destination_path,
            overwrite_existing=False,
            delete_on_failure=True,
        )
        self.assertFalse(result)

    def test_download_object_success_not_found(self):
        CloudFilesMockHttp.type = "NOT_FOUND"
        container = Container(name="foo_bar_container", extra={}, driver=self)

        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=CloudFilesStorageDriver,
        )
        destination_path = os.path.abspath(__file__) + ".temp"
        try:
            self.driver.download_object(
                obj=obj,
                destination_path=destination_path,
                overwrite_existing=False,
                delete_on_failure=True,
            )
        except ObjectDoesNotExistError:
            pass
        else:
            self.fail("Object does not exist but an exception was not thrown")

    def test_download_object_range_success(self):
        container = Container(name="foo_bar_container", extra={}, driver=self)
        obj = Object(
            name="foo_bar_object_range",
            size=10,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=CloudFilesStorageDriver,
        )
        destination_path = os.path.abspath(__file__) + ".temp"
        result = self.driver.download_object_range(
            obj=obj,
            destination_path=destination_path,
            start_bytes=5,
            end_bytes=7,
            overwrite_existing=False,
            delete_on_failure=True,
        )
        self.assertTrue(result)

        with open(destination_path) as fp:
            content = fp.read()

        self.assertEqual(content, "56")

    def test_download_object_as_stream(self):
        container = Container(name="foo_bar_container", extra={}, driver=self)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=CloudFilesStorageDriver,
        )

        stream = self.driver.download_object_as_stream(obj=obj, chunk_size=None)
        self.assertTrue(hasattr(stream, "__iter__"))

    def test_download_object_as_stream_data_is_not_buffered_in_memory(self):
        # Test case which verifies that response.response attribute is not accessed
        # and as such, whole body response is not buffered into RAM

        # If content is consumed and response.content attribute accessed exception
        # will be thrown and test will fail
        mock_response = Mock(name="mock response")
        mock_response.headers = {}
        mock_response.status = 200
        msg1 = '"response" attribute was accessed but it shouldn\'t have been'
        msg2 = '"content" attribute was accessed but it shouldn\'t have been'
        type(mock_response).response = PropertyMock(
            name="mock response attribute", side_effect=Exception(msg1)
        )
        type(mock_response).content = PropertyMock(
            name="mock content attribute", side_effect=Exception(msg2)
        )
        mock_response.iter_content.return_value = StringIO("a" * 1000)

        self.driver.connection.request = Mock()
        self.driver.connection.request.return_value = mock_response

        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object_NO_BUFFER",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver,
        )
        result = self.driver.download_object_as_stream(obj=obj)
        result = exhaust_iterator(result)
        result = result.decode("utf-8")

        self.assertEqual(result, "a" * 1000)

    def test_download_object_range_as_stream_success(self):
        container = Container(name="foo_bar_container", extra={}, driver=self)
        obj = Object(
            name="foo_bar_object_range",
            size=2,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=CloudFilesStorageDriver,
        )

        stream = self.driver.download_object_range_as_stream(
            start_bytes=5, end_bytes=7, obj=obj, chunk_size=None
        )
        self.assertTrue(hasattr(stream, "__iter__"))
        consumed_stream = "".join(chunk.decode("utf-8") for chunk in stream)
        self.assertEqual(consumed_stream, "56")
        self.assertEqual(len(consumed_stream), obj.size)

    def test_upload_object_success(self):
        def upload_file(
            self,
            object_name=None,
            content_type=None,
            request_path=None,
            request_method=None,
            headers=None,
            file_path=None,
            stream=None,
        ):
            return {
                "response": make_response(
                    201, headers={"etag": "0cc175b9c0f1b6a831c399e269772661"}
                ),
                "bytes_transferred": 1000,
                "data_hash": "0cc175b9c0f1b6a831c399e269772661",
            }

        old_func = CloudFilesStorageDriver._upload_object
        CloudFilesStorageDriver._upload_object = upload_file
        file_path = os.path.abspath(__file__)
        container = Container(name="foo_bar_container", extra={}, driver=self)
        object_name = "foo_test_upload"
        extra = {"meta_data": {"some-value": "foobar"}}
        obj = self.driver.upload_object(
            file_path=file_path,
            container=container,
            extra=extra,
            object_name=object_name,
        )
        self.assertEqual(obj.name, "foo_test_upload")
        self.assertEqual(obj.size, 1000)
        self.assertTrue("some-value" in obj.meta_data)
        CloudFilesStorageDriver._upload_object = old_func

    def test_upload_object_zero_size_object(self):
        def upload_file(
            self,
            object_name=None,
            content_type=None,
            request_path=None,
            request_method=None,
            headers=None,
            file_path=None,
            stream=None,
        ):
            return {
                "response": make_response(
                    201, headers={"etag": "0cc175b9c0f1b6a831c399e269772661"}
                ),
                "bytes_transferred": 0,
                "data_hash": "0cc175b9c0f1b6a831c399e269772661",
            }

        old_func = CloudFilesStorageDriver._upload_object
        CloudFilesStorageDriver._upload_object = upload_file

        old_request = self.driver.connection.request

        file_path = os.path.join(os.path.dirname(__file__), "__init__.py")
        container = Container(name="foo_bar_container", extra={}, driver=self)
        object_name = "empty"
        extra = {}

        def func(*args, **kwargs):
            self.assertEqual(kwargs["headers"]["Content-Length"], 0)
            func.called = True
            return old_request(*args, **kwargs)

        self.driver.connection.request = func
        func.called = False
        obj = self.driver.upload_object(
            file_path=file_path,
            container=container,
            extra=extra,
            object_name=object_name,
        )
        self.assertEqual(obj.name, "empty")
        self.assertEqual(obj.size, 0)
        CloudFilesStorageDriver._upload_object = old_func
        self.driver.connection.request = old_request

    def test_upload_object_invalid_hash(self):
        CloudFilesMockHttp.type = "INVALID_HASH"

        def upload_file(
            self,
            object_name=None,
            content_type=None,
            request_path=None,
            request_method=None,
            headers=None,
            file_path=None,
            stream=None,
        ):
            return {
                "response": make_response(
                    201, headers={"etag": "0cc175b9c0f1b6a831c399e269772661"}
                ),
                "bytes_transferred": 1000,
                "data_hash": "blah blah",
            }

        old_func = CloudFilesStorageDriver._upload_object
        CloudFilesStorageDriver._upload_object = upload_file
        file_path = os.path.abspath(__file__)
        container = Container(name="foo_bar_container", extra={}, driver=self)
        object_name = "foo_test_upload"
        try:
            self.driver.upload_object(
                file_path=file_path,
                container=container,
                object_name=object_name,
                verify_hash=True,
            )
        except ObjectHashMismatchError:
            pass
        else:
            self.fail("Invalid hash was returned but an exception was not thrown")
        finally:
            CloudFilesStorageDriver._upload_object = old_func

    def test_upload_object_no_content_type(self):
        def no_content_type(name):
            return None, None

        old_func = libcloud.utils.files.guess_file_mime_type
        libcloud.utils.files.guess_file_mime_type = no_content_type
        file_path = os.path.abspath(__file__)
        container = Container(name="foo_bar_container", extra={}, driver=self)
        object_name = "foo_test_upload"

        obj = self.driver.upload_object(
            file_path=file_path,
            verify_hash=False,
            container=container,
            object_name=object_name,
        )

        self.assertEqual(obj.name, object_name)
        libcloud.utils.files.guess_file_mime_type = old_func

    def test_upload_object_inexistent_file(self):
        def dummy_content_type(name):
            return "application/zip", None

        old_func = libcloud.utils.files.guess_file_mime_type
        libcloud.utils.files.guess_file_mime_type = dummy_content_type

        file_path = os.path.abspath(__file__ + ".inexistent")
        container = Container(name="foo_bar_container", extra={}, driver=self)
        object_name = "foo_test_upload"
        try:
            self.driver.upload_object(
                file_path=file_path, container=container, object_name=object_name
            )
        except OSError:
            pass
        else:
            self.fail("Inexistent but an exception was not thrown")
        finally:
            libcloud.utils.files.guess_file_mime_type = old_func

    def test_upload_object_via_stream(self):
        def dummy_content_type(name):
            return "application/zip", None

        old_func = libcloud.utils.files.guess_file_mime_type
        libcloud.utils.files.guess_file_mime_type = dummy_content_type

        container = Container(name="foo_bar_container", extra={}, driver=self)
        object_name = "foo_test_stream_data"
        iterator = BytesIO(b("235"))
        try:
            self.driver.upload_object_via_stream(
                container=container, object_name=object_name, iterator=iterator
            )
        finally:
            libcloud.utils.files.guess_file_mime_type = old_func

    def test_upload_object_via_stream_stream_seek_at_end(self):
        def dummy_content_type(name):
            return "application/zip", None

        old_func = libcloud.utils.files.guess_file_mime_type
        libcloud.utils.files.guess_file_mime_type = dummy_content_type

        container = Container(name="foo_bar_container", extra={}, driver=self)
        object_name = "foo_test_stream_data_seek"
        iterator = BytesIO(b("123456789"))
        iterator.seek(10)

        self.assertEqual(iterator.tell(), 10)

        try:
            self.driver.upload_object_via_stream(
                container=container, object_name=object_name, iterator=iterator
            )
        finally:
            libcloud.utils.files.guess_file_mime_type = old_func

    def test_delete_object_success(self):
        container = Container(name="foo_bar_container", extra={}, driver=self)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=CloudFilesStorageDriver,
        )
        status = self.driver.delete_object(obj=obj)
        self.assertTrue(status)

    def test_delete_object_not_found(self):
        CloudFilesMockHttp.type = "NOT_FOUND"
        container = Container(name="foo_bar_container", extra={}, driver=self)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=CloudFilesStorageDriver,
        )
        try:
            self.driver.delete_object(obj=obj)
        except ObjectDoesNotExistError:
            pass
        else:
            self.fail("Object does not exist but an exception was not thrown")

    def test_ex_get_meta_data(self):
        meta_data = self.driver.ex_get_meta_data()
        self.assertTrue(isinstance(meta_data, dict))
        self.assertTrue("object_count" in meta_data)
        self.assertTrue("container_count" in meta_data)
        self.assertTrue("bytes_used" in meta_data)
        self.assertTrue("temp_url_key" in meta_data)

    def test_ex_purge_object_from_cdn(self):
        CloudFilesMockHttp.type = "PURGE_SUCCESS"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self,
        )

        self.assertTrue(self.driver.ex_purge_object_from_cdn(obj=obj))

    def test_ex_purge_object_from_cdn_with_email(self):
        CloudFilesMockHttp.type = "PURGE_SUCCESS_EMAIL"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self,
        )
        self.assertTrue(self.driver.ex_purge_object_from_cdn(obj=obj, email="test@test.com"))

    @mock.patch("os.path.getsize")
    def test_ex_multipart_upload_object_for_small_files(self, getsize_mock):
        getsize_mock.return_value = 0

        old_func = CloudFilesStorageDriver.upload_object
        mocked_upload_object = mock.Mock(return_value="test")
        CloudFilesStorageDriver.upload_object = mocked_upload_object

        file_path = os.path.abspath(__file__)
        container = Container(name="foo_bar_container", extra={}, driver=self)
        object_name = "foo_test_upload"
        obj = self.driver.ex_multipart_upload_object(
            file_path=file_path, container=container, object_name=object_name
        )
        CloudFilesStorageDriver.upload_object = old_func

        self.assertTrue(mocked_upload_object.called)
        self.assertEqual(obj, "test")

    def test_ex_multipart_upload_object_success(self):
        _upload_object_part = CloudFilesStorageDriver._upload_object_part
        _upload_object_manifest = CloudFilesStorageDriver._upload_object_manifest

        mocked__upload_object_part = mock.Mock(return_value="test_part")
        mocked__upload_object_manifest = mock.Mock(return_value="test_manifest")

        CloudFilesStorageDriver._upload_object_part = mocked__upload_object_part
        CloudFilesStorageDriver._upload_object_manifest = mocked__upload_object_manifest

        parts = 5
        file_path = os.path.abspath(__file__)
        chunk_size = int(math.ceil(float(os.path.getsize(file_path)) / parts))
        container = Container(name="foo_bar_container", extra={}, driver=self)
        object_name = "foo_test_upload"
        self.driver.ex_multipart_upload_object(
            file_path=file_path,
            container=container,
            object_name=object_name,
            chunk_size=chunk_size,
        )

        CloudFilesStorageDriver._upload_object_part = _upload_object_part
        CloudFilesStorageDriver._upload_object_manifest = _upload_object_manifest

        self.assertEqual(mocked__upload_object_part.call_count, parts)
        self.assertTrue(mocked__upload_object_manifest.call_count, 1)

    def test__upload_object_part(self):
        _put_object = CloudFilesStorageDriver._put_object
        mocked__put_object = mock.Mock(return_value="test")
        CloudFilesStorageDriver._put_object = mocked__put_object

        part_number = 7
        object_name = "test_object"
        expected_name = object_name + "/%08d" % part_number
        container = Container(name="foo_bar_container", extra={}, driver=self)

        self.driver._upload_object_part(container, object_name, part_number, None)

        CloudFilesStorageDriver._put_object = _put_object

        func_kwargs = tuple(mocked__put_object.call_args)[1]
        self.assertEqual(func_kwargs["object_name"], expected_name)
        self.assertEqual(func_kwargs["container"], container)

    def test_upload_object_via_stream_with_cors_headers(self):
        """
        Test we can add some ``Cross-origin resource sharing`` headers
        to the request about to be sent.
        """
        cors_headers = {
            "Access-Control-Allow-Origin": "http://mozilla.com",
            "Origin": "http://storage.clouddrive.com",
        }
        expected_headers = {
            # Automatically added headers
            "Content-Type": "application/octet-stream"
        }
        expected_headers.update(cors_headers)

        def intercept_request(request_path, method=None, data=None, headers=None, raw=True):
            # What we're actually testing
            self.assertDictEqual(expected_headers, headers)

            raise NotImplementedError("oops")

        self.driver.connection.request = intercept_request

        container = Container(name="CORS", extra={}, driver=self.driver)

        try:
            self.driver.upload_object_via_stream(
                iterator=iter(b"blob data like an image or video"),
                container=container,
                object_name="test_object",
                headers=cors_headers,
            )
        except NotImplementedError:
            # Don't care about the response we'd have to mock anyway
            # as long as we intercepted the request and checked its headers
            pass
        else:
            self.fail(
                "Expected NotImplementedError to be thrown to "
                "verify we actually checked the expected headers"
            )

    def test_upload_object_via_stream_python3_bytes_error(self):
        container = Container(name="py3", extra={}, driver=self.driver)
        bytes_blob = b"blob data like an image or video"

        # This is mostly to check we didn't discover other errors along the way
        mocked_response = container.upload_object_via_stream(
            iterator=iter(bytes_blob),
            object_name="img_or_vid",
        )
        self.assertEqual(len(bytes_blob), mocked_response.size)

    @unittest.skip("Skipping as chunking is disabled in 2.0rc1")
    def test_upload_object_via_stream_chunked_encoding(self):
        # Create enough bytes it should get split into two chunks
        bytes_blob = "".join(["\0" for _ in range(CHUNK_SIZE + 1)])
        hex_chunk_size = ("%X" % CHUNK_SIZE).encode("utf8")
        expected = [
            # Chunk 1
            hex_chunk_size + b"\r\n",
            bytes(bytes_blob[:CHUNK_SIZE].encode("utf8")),
            b"\r\n",
            # Chunk 2
            b"1\r\n",
            bytes(bytes_blob[CHUNK_SIZE:].encode("utf8")),
            b"\r\n",
            # If chunked, also send a final message
            b"0\r\n\r\n",
        ]
        logged_data = []

        class InterceptResponse(MockHttp):
            def __init__(self, connection, response=None):
                super().__init__(connection=connection, response=response)
                old_send = self.connection.connection.send

                def intercept_send(data):
                    old_send(data)
                    logged_data.append(data)

                self.connection.connection.send = intercept_send

            def _v1_MossoCloudFS_py3_img_or_vid2(self, method, url, body, headers):
                headers = {"etag": "d79fb00c27b50494a463e680d459c90c"}
                headers.update(self.base_headers)
                _201 = httplib.CREATED
                return _201, "", headers, httplib.responses[_201]

        self.driver_klass.connectionCls.rawResponseCls = InterceptResponse

        container = Container(name="py3", extra={}, driver=self.driver)
        container.upload_object_via_stream(
            iterator=iter(bytes_blob),
            object_name="img_or_vid2",
        )
        self.assertListEqual(expected, logged_data)

    def test__upload_object_manifest(self):
        hash_function = self.driver._get_hash_function()
        hash_function.update(b(""))
        data_hash = hash_function.hexdigest()

        fake_response = type("CloudFilesResponse", (), {"headers": {"etag": data_hash}})

        _request = self.driver.connection.request
        mocked_request = mock.Mock(return_value=fake_response)
        self.driver.connection.request = mocked_request

        container = Container(name="foo_bar_container", extra={}, driver=self)
        object_name = "test_object"

        self.driver._upload_object_manifest(container, object_name)

        func_args, func_kwargs = tuple(mocked_request.call_args)

        self.driver.connection.request = _request

        self.assertEqual(func_args[0], "/" + container.name + "/" + object_name)
        self.assertEqual(
            func_kwargs["headers"]["X-Object-Manifest"],
            container.name + "/" + object_name + "/",
        )
        self.assertEqual(func_kwargs["method"], "PUT")

    def test__upload_object_manifest_wrong_hash(self):
        fake_response = type("CloudFilesResponse", (), {"headers": {"etag": "0000000"}})

        _request = self.driver.connection.request
        mocked_request = mock.Mock(return_value=fake_response)
        self.driver.connection.request = mocked_request

        container = Container(name="foo_bar_container", extra={}, driver=self)
        object_name = "test_object"

        try:
            self.driver._upload_object_manifest(container, object_name)
        except ObjectHashMismatchError:
            pass
        else:
            self.fail("Exception was not thrown")
        finally:
            self.driver.connection.request = _request

    def test_create_container_put_object_name_encoding(self):
        def upload_file(
            self,
            object_name=None,
            content_type=None,
            request_path=None,
            request_method=None,
            headers=None,
            file_path=None,
            stream=None,
        ):
            return {
                "response": make_response(
                    201, headers={"etag": "0cc175b9c0f1b6a831c399e269772661"}
                ),
                "bytes_transferred": 1000,
                "data_hash": "0cc175b9c0f1b6a831c399e269772661",
            }

        old_func = CloudFilesStorageDriver._upload_object
        CloudFilesStorageDriver._upload_object = upload_file

        container_name = "speci@l_name"
        object_name = "m@obj€ct"
        file_path = os.path.abspath(__file__)

        container = self.driver.create_container(container_name=container_name)
        self.assertEqual(container.name, container_name)

        obj = self.driver.upload_object(
            file_path=file_path, container=container, object_name=object_name
        )
        self.assertEqual(obj.name, object_name)
        CloudFilesStorageDriver._upload_object = old_func

    def test_ex_enable_static_website(self):
        container = Container(name="foo_bar_container", extra={}, driver=self)
        result = self.driver.ex_enable_static_website(container=container, index_file="index.html")
        self.assertTrue(result)

    def test_ex_set_error_page(self):
        container = Container(name="foo_bar_container", extra={}, driver=self)
        result = self.driver.ex_set_error_page(container=container, file_name="error.html")
        self.assertTrue(result)

    def test_ex_set_account_metadata_temp_url_key(self):
        result = self.driver.ex_set_account_metadata_temp_url_key("a key")
        self.assertTrue(result)

    @mock.patch("libcloud.storage.drivers.cloudfiles.time")
    def test_ex_get_object_temp_url(self, time):
        time.return_value = 0
        self.driver.ex_get_meta_data = mock.Mock()
        self.driver.ex_get_meta_data.return_value = {
            "container_count": 1,
            "object_count": 1,
            "bytes_used": 1,
            "temp_url_key": "foo",
        }
        container = Container(name="foo_bar_container", extra={}, driver=self)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self,
        )
        hmac_body = "{}\n{}\n{}".format(
            "GET",
            60,
            "/v1/MossoCloudFS/foo_bar_container/foo_bar_object",
        )
        sig = hmac.new(b("foo"), b(hmac_body), sha1).hexdigest()
        ret = self.driver.ex_get_object_temp_url(obj, "GET")
        temp_url = (
            "https://storage4.%s1.clouddrive.com/v1/MossoCloudFS/"
            "foo_bar_container/foo_bar_object?temp_url_expires=60&temp_url_sig=%s"
            % (self.region, sig)
        )

        self.assertEqual("".join(sorted(ret)), "".join(sorted(temp_url)))

    def test_ex_get_object_temp_url_no_key_raises_key_error(self):
        self.driver.ex_get_meta_data = mock.Mock()
        self.driver.ex_get_meta_data.return_value = {
            "container_count": 1,
            "object_count": 1,
            "bytes_used": 1,
            "temp_url_key": None,
        }
        container = Container(name="foo_bar_container", extra={}, driver=self)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self,
        )
        self.assertRaises(KeyError, self.driver.ex_get_object_temp_url, obj, "GET")

    def _remove_test_file(self):
        file_path = os.path.abspath(__file__) + ".temp"

        try:
            os.unlink(file_path)
        except OSError:
            pass


class CloudFilesDeprecatedUSTests(CloudFilesTests):
    driver_klass = CloudFilesStorageDriver
    region = "ord"


class CloudFilesDeprecatedUKTests(CloudFilesTests):
    driver_klass = CloudFilesStorageDriver
    region = "lon"


class CloudFilesMockHttp(BaseRangeDownloadMockHttp, unittest.TestCase):
    fixtures = StorageFileFixtures("cloudfiles")
    base_headers = {"content-type": "application/json; charset=UTF-8"}

    # fake auth token response
    def _v2_0_tokens(self, method, url, body, headers):
        headers = copy.deepcopy(self.base_headers)
        body = self.fixtures.load("_v2_0__auth.json")
        return (httplib.OK, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_MALFORMED_JSON(self, method, url, body, headers):
        # test_invalid_json_throws_exception
        body = 'broken: json /*"'
        return (
            httplib.NO_CONTENT,
            body,
            self.base_headers,
            httplib.responses[httplib.OK],
        )

    def _v1_MossoCloudFS_EMPTY(self, method, url, body, headers):
        return (
            httplib.NO_CONTENT,
            body,
            self.base_headers,
            httplib.responses[httplib.OK],
        )

    def _v1_MossoCloudFS(self, method, url, body, headers):
        headers = copy.deepcopy(self.base_headers)
        if method == "GET":
            # list_containers
            body = self.fixtures.load("list_containers.json")
            status_code = httplib.OK
        elif method == "HEAD":
            # get_meta_data
            body = self.fixtures.load("meta_data.json")
            status_code = httplib.NO_CONTENT
            headers.update(
                {
                    "x-account-container-count": "10",
                    "x-account-object-count": "400",
                    "x-account-bytes-used": "1234567",
                }
            )
        elif method == "POST":
            body = ""
            status_code = httplib.NO_CONTENT
        return (status_code, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_not_found(self, method, url, body, headers):
        # test_get_object_not_found
        if method == "HEAD":
            body = ""
        else:
            raise ValueError("Invalid method")

        return (
            httplib.NOT_FOUND,
            body,
            self.base_headers,
            httplib.responses[httplib.OK],
        )

    def _v1_MossoCloudFS_test_container_EMPTY(self, method, url, body, headers):
        body = self.fixtures.load("list_container_objects_empty.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_test_20container_201_EMPTY(self, method, url, body, headers):
        body = self.fixtures.load("list_container_objects_empty.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_test_container(self, method, url, body, headers):
        headers = copy.deepcopy(self.base_headers)
        if method == "GET":
            # list_container_objects
            if url.find("marker") == -1:
                body = self.fixtures.load("list_container_objects.json")
                status_code = httplib.OK
            else:
                body = ""
                status_code = httplib.NO_CONTENT
        elif method == "HEAD":
            # get_container
            body = self.fixtures.load("list_container_objects_empty.json")
            status_code = httplib.NO_CONTENT
            headers.update({"x-container-object-count": "800", "x-container-bytes-used": "1234568"})
        return (status_code, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_test_container_ITERATOR(self, method, url, body, headers):
        headers = copy.deepcopy(self.base_headers)
        # list_container_objects
        if url.find("foo-test-3") != -1:
            body = self.fixtures.load("list_container_objects_not_exhausted2.json")
            status_code = httplib.OK
        elif url.find("foo-test-5") != -1:
            body = ""
            status_code = httplib.NO_CONTENT
        else:
            # First request
            body = self.fixtures.load("list_container_objects_not_exhausted1.json")
            status_code = httplib.OK

        return (status_code, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_test_container_not_found(self, method, url, body, headers):
        # test_get_container_not_found
        if method == "HEAD":
            body = ""
        else:
            raise ValueError("Invalid method")

        return (
            httplib.NOT_FOUND,
            body,
            self.base_headers,
            httplib.responses[httplib.OK],
        )

    def _v1_MossoCloudFS_test_container_test_object(self, method, url, body, headers):
        headers = copy.deepcopy(self.base_headers)
        if method == "HEAD":
            # get_object
            body = self.fixtures.load("list_container_objects_empty.json")
            status_code = httplib.NO_CONTENT
            headers.update(
                {
                    "content-length": "555",
                    "last-modified": "Tue, 25 Jan 2011 22:01:49 GMT",
                    "etag": "6b21c4a111ac178feacf9ec9d0c71f17",
                    "x-object-meta-foo-bar": "test 1",
                    "x-object-meta-bar-foo": "test 2",
                    "content-type": "application/zip",
                }
            )
        return (status_code, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_test_container__7E_test_object(self, method, url, body, headers):
        headers = copy.deepcopy(self.base_headers)
        if method == "HEAD":
            # get_object_name_encoding
            body = self.fixtures.load("list_container_objects_empty.json")
            status_code = httplib.NO_CONTENT
            headers.update(
                {
                    "content-length": "555",
                    "last-modified": "Tue, 25 Jan 2011 22:01:49 GMT",
                    "etag": "6b21c4a111ac178feacf9ec9d0c71f17",
                    "x-object-meta-foo-bar": "test 1",
                    "x-object-meta-bar-foo": "test 2",
                    "content-type": "application/zip",
                }
            )
        return (status_code, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_test_create_container(self, method, url, body, headers):
        # test_create_container_success
        headers = copy.deepcopy(self.base_headers)
        body = self.fixtures.load("list_container_objects_empty.json")
        headers = copy.deepcopy(self.base_headers)
        headers.update({"content-length": "18", "date": "Mon, 28 Feb 2011 07:52:57 GMT"})
        status_code = httplib.CREATED
        return (status_code, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_speci_40l_name(self, method, url, body, headers):
        # test_create_container_put_object_name_encoding
        # Verify that the name is properly url encoded
        container_name = "speci@l_name"
        encoded_container_name = urlquote(container_name)
        self.assertTrue(encoded_container_name in url)

        headers = copy.deepcopy(self.base_headers)
        body = self.fixtures.load("list_container_objects_empty.json")
        headers = copy.deepcopy(self.base_headers)
        headers.update({"content-length": "18", "date": "Mon, 28 Feb 2011 07:52:57 GMT"})
        status_code = httplib.CREATED
        return (status_code, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_test_create_container_ALREADY_EXISTS(self, method, url, body, headers):
        # test_create_container_already_exists
        headers = copy.deepcopy(self.base_headers)
        body = self.fixtures.load("list_container_objects_empty.json")
        headers.update({"content-type": "text/plain"})
        status_code = httplib.ACCEPTED
        return (status_code, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_foo_bar_container(self, method, url, body, headers):
        if method == "DELETE":
            # test_delete_container_success
            body = self.fixtures.load("list_container_objects_empty.json")
            headers = self.base_headers
            status_code = httplib.NO_CONTENT
        elif method == "POST":
            # test_ex_enable_static_website
            body = ""
            headers = self.base_headers
            status_code = httplib.ACCEPTED
        return (status_code, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_foo_bar_container_object_PURGE_SUCCESS(self, method, url, body, headers):
        if method == "DELETE":
            # test_ex_purge_from_cdn
            headers = self.base_headers
            status_code = httplib.NO_CONTENT
        return (status_code, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_foo_bar_container_object_PURGE_SUCCESS_EMAIL(
        self, method, url, body, headers
    ):
        if method == "DELETE":
            # test_ex_purge_from_cdn_with_email
            self.assertEqual(headers["X-Purge-Email"], "test@test.com")
            headers = self.base_headers
            status_code = httplib.NO_CONTENT
        return (status_code, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_foo_bar_container_NOT_FOUND(self, method, url, body, headers):
        if method == "DELETE":
            # test_delete_container_not_found
            body = self.fixtures.load("list_container_objects_empty.json")
            headers = self.base_headers
            status_code = httplib.NOT_FOUND
        return (status_code, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_foo_bar_container_NOT_EMPTY(self, method, url, body, headers):
        if method == "DELETE":
            # test_delete_container_not_empty
            body = self.fixtures.load("list_container_objects_empty.json")
            headers = self.base_headers
            status_code = httplib.CONFLICT
        return (status_code, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_foo_bar_container_foo_bar_object(self, method, url, body, headers):
        if method == "DELETE":
            # test_delete_object_success
            body = self.fixtures.load("list_container_objects_empty.json")
            headers = self.base_headers
            status_code = httplib.NO_CONTENT
            return (status_code, body, headers, httplib.responses[httplib.OK])
        elif method == "GET":
            body = generate_random_data(1000)
            return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_foo_bar_container_foo_bar_object_range(self, method, url, body, headers):
        if method == "GET":
            # test_download_object_range_success
            body = "0123456789123456789"

            self.assertTrue("Range" in headers)
            self.assertEqual(headers["Range"], "bytes=5-6")

            start_bytes, end_bytes = self._get_start_and_end_bytes_from_range_str(
                headers["Range"], body
            )

            return (
                httplib.PARTIAL_CONTENT,
                body[start_bytes : end_bytes + 1],
                self.base_headers,
                httplib.responses[httplib.PARTIAL_CONTENT],
            )

    def _v1_MossoCloudFS_py3_img_or_vid(self, method, url, body, headers):
        headers = {"etag": "e2378cace8712661ce7beec3d9362ef6"}
        headers.update(self.base_headers)
        return httplib.CREATED, "", headers, httplib.responses[httplib.CREATED]

    def _v1_MossoCloudFS_foo_bar_container_foo_test_upload(self, method, url, body, headers):
        # test_object_upload_success

        body = ""
        headers = {}
        headers.update(self.base_headers)
        headers["etag"] = "hash343hhash89h932439jsaa89"
        return (httplib.CREATED, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_speci_40l_name_m_40obj_E2_82_ACct(self, method, url, body, headers):
        # test_create_container_put_object_name_encoding
        # Verify that the name is properly url encoded
        object_name = "m@obj€ct"
        urlquote(object_name)

        headers = copy.deepcopy(self.base_headers)
        body = ""
        headers["etag"] = "hash343hhash89h932439jsaa89"
        return (httplib.CREATED, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_foo_bar_container_empty(self, method, url, body, headers):
        # test_upload_object_zero_size_object
        body = ""
        headers = {}
        headers.update(self.base_headers)
        headers["etag"] = "hash343hhash89h932439jsaa89"
        return (httplib.CREATED, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_foo_bar_container_foo_test_upload_INVALID_HASH(
        self, method, url, body, headers
    ):
        # test_object_upload_invalid_hash
        body = ""
        headers = {}
        headers.update(self.base_headers)
        headers["etag"] = "foobar"
        return (httplib.CREATED, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_foo_bar_container_foo_bar_object_INVALID_SIZE(
        self, method, url, body, headers
    ):
        # test_download_object_invalid_file_size
        body = generate_random_data(100)
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_foo_bar_container_foo_bar_object_NOT_FOUND(
        self, method, url, body, headers
    ):
        body = ""
        return (
            httplib.NOT_FOUND,
            body,
            self.base_headers,
            httplib.responses[httplib.OK],
        )

    def _v1_MossoCloudFS_foo_bar_container_foo_test_stream_data(self, method, url, body, headers):
        # test_upload_object_via_stream_success
        hasher = hashlib.md5()
        hasher.update(b"235")
        hash_value = hasher.hexdigest()

        headers = {}
        headers.update(self.base_headers)
        headers["etag"] = hash_value
        body = "test"
        return (httplib.CREATED, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_foo_bar_container_foo_test_stream_data_seek(
        self, method, url, body, headers
    ):
        # test_upload_object_via_stream_stream_seek_at_end
        hasher = hashlib.md5()
        hasher.update(b"123456789")
        hash_value = hasher.hexdigest()

        headers = {}
        headers.update(self.base_headers)
        headers["etag"] = hash_value
        body = "test"
        return (httplib.CREATED, body, headers, httplib.responses[httplib.OK])

    def _v1_MossoCloudFS_foo_bar_container_foo_bar_object_NO_BUFFER(
        self, method, url, body, headers
    ):
        # test_download_object_data_is_not_buffered_in_memory
        headers = {}
        headers.update(self.base_headers)
        headers["etag"] = "577ef1154f3240ad5b9b413aa7346a1e"
        body = generate_random_data(1000)
        return (httplib.OK, body, headers, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from libcloud.utils.py3 import httplib
from libcloud.utils.misc import reverse_dict
from libcloud.common.brightbox import BrightboxConnection
from libcloud.loadbalancer.base import Driver, Member, Algorithm, LoadBalancer
from libcloud.loadbalancer.types import State

API_VERSION = "1.0"


class BrightboxLBDriver(Driver):
    connectionCls = BrightboxConnection

    name = "Brightbox"
    website = "http://www.brightbox.co.uk/"

    LB_STATE_MAP = {
        "creating": State.PENDING,
        "active": State.RUNNING,
        "deleting": State.UNKNOWN,
        "deleted": State.UNKNOWN,
        "failing": State.UNKNOWN,
        "failed": State.UNKNOWN,
    }

    _VALUE_TO_ALGORITHM_MAP = {
        "round-robin": Algorithm.ROUND_ROBIN,
        "least-connections": Algorithm.LEAST_CONNECTIONS,
    }

    _ALGORITHM_TO_VALUE_MAP = reverse_dict(_VALUE_TO_ALGORITHM_MAP)

    def list_protocols(self):
        return ["tcp", "http"]

    def list_balancers(self):
        data = self.connection.request("/%s/load_balancers" % API_VERSION).object

        return list(map(self._to_balancer, data))

    def create_balancer(self, name, port, protocol, algorithm, members):
        response = self._post(
            "/%s/load_balancers" % API_VERSION,
            {
                "name": name,
                "nodes": list(map(self._member_to_node, members)),
                "policy": self._algorithm_to_value(algorithm),
                "listeners": [{"in": port, "out": port, "protocol": protocol}],
                "healthcheck": {"type": protocol, "port": port},
            },
        )

        return self._to_balancer(response.object)

    def destroy_balancer(self, balancer):
        response = self.connection.request(
            "/{}/load_balancers/{}".format(API_VERSION, balancer.id), method="DELETE"
        )

        return response.status == httplib.ACCEPTED

    def get_balancer(self, balancer_id):
        data = self.connection.request(
            "/{}/load_balancers/{}".format(API_VERSION, balancer_id)
        ).object
        return self._to_balancer(data)

    def balancer_attach_compute_node(self, balancer, node):
        return self.balancer_attach_member(balancer, node)

    def balancer_attach_member(self, balancer, member):
        path = "/{}/load_balancers/{}/add_nodes".format(API_VERSION, balancer.id)

        self._post(path, {"nodes": [self._member_to_node(member)]})

        return member

    def balancer_detach_member(self, balancer, member):
        path = "/{}/load_balancers/{}/remove_nodes".format(API_VERSION, balancer.id)

        response = self._post(path, {"nodes": [self._member_to_node(member)]})

        return response.status == httplib.ACCEPTED

    def balancer_list_members(self, balancer):
        path = "/{}/load_balancers/{}".format(API_VERSION, balancer.id)

        data = self.connection.request(path).object

        def func(data):
            return self._node_to_member(data, balancer)

        return list(map(func, data["nodes"]))

    def _post(self, path, data={}):
        headers = {"Content-Type": "application/json"}

        return self.connection.request(path, data=data, headers=headers, method="POST")

    def _to_balancer(self, data):
        return LoadBalancer(
            id=data["id"],
            name=data["name"],
            state=self.LB_STATE_MAP.get(data["status"], State.UNKNOWN),
            ip=self._public_ip(data),
            port=data["listeners"][0]["in"],
            driver=self.connection.driver,
        )

    def _member_to_node(self, member):
        return {"node": member.id}

    def _node_to_member(self, data, balancer):
        return Member(id=data["id"], ip=None, port=None, balancer=balancer)

    def _public_ip(self, data):
        if len(data["cloud_ips"]) > 0:
            ip = data["cloud_ips"][0]["public_ip"]
        else:
            ip = None

        return ip



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.common.aws import SignedAWSConnection
from libcloud.common.types import LibcloudError
from libcloud.storage.drivers.s3 import S3Connection, BaseS3Connection, BaseS3StorageDriver

__all__ = ["DigitalOceanSpacesStorageDriver"]

DO_SPACES_HOSTS_BY_REGION = {
    "nyc3": "nyc3.digitaloceanspaces.com",
    "ams3": "ams3.digitaloceanspaces.com",
    "sfo2": "sfo2.digitaloceanspaces.com",
    "sgp1": "sgp1.digitaloceanspaces.com",
}
DO_SPACES_DEFAULT_REGION = "nyc3"
DEFAULT_SIGNATURE_VERSION = "2"
S3_API_VERSION = "2006-03-01"


class DOSpacesConnectionAWS4(SignedAWSConnection, BaseS3Connection):
    service_name = "s3"
    version = S3_API_VERSION

    def __init__(
        self,
        user_id,
        key,
        secure=True,
        host=None,
        port=None,
        url=None,
        timeout=None,
        proxy_url=None,
        token=None,
        retry_delay=None,
        backoff=None,
        **kwargs,
    ):
        super().__init__(
            user_id,
            key,
            secure,
            host,
            port,
            url,
            timeout,
            proxy_url,
            token,
            retry_delay,
            backoff,
            signature_version=4,
        )


class DOSpacesConnectionAWS2(S3Connection):
    def __init__(
        self,
        user_id,
        key,
        secure=True,
        host=None,
        port=None,
        url=None,
        timeout=None,
        proxy_url=None,
        token=None,
        retry_delay=None,
        backoff=None,
        **kwargs,
    ):
        super().__init__(
            user_id,
            key,
            secure,
            host,
            port,
            url,
            timeout,
            proxy_url,
            token,
            retry_delay,
            backoff,
        )


class DigitalOceanSpacesStorageDriver(BaseS3StorageDriver):
    name = "DigitalOcean Spaces"
    website = "https://www.digitalocean.com/products/object-storage/"
    supports_chunked_encoding = False
    supports_s3_multipart_upload = True

    def __init__(
        self,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        api_version=None,
        region=DO_SPACES_DEFAULT_REGION,
        **kwargs,
    ):
        if region not in DO_SPACES_HOSTS_BY_REGION:
            raise LibcloudError("Unknown region (%s)" % (region), driver=self)

        host = DO_SPACES_HOSTS_BY_REGION[region]
        self.name = "DigitalOcean Spaces (%s)" % (region)

        self.region_name = region
        self.signature_version = str(kwargs.pop("signature_version", DEFAULT_SIGNATURE_VERSION))

        if self.signature_version == "2":
            self.connectionCls = DOSpacesConnectionAWS2
        elif self.signature_version == "4":
            self.connectionCls = DOSpacesConnectionAWS4
        else:
            raise ValueError("Invalid signature_version: %s" % (self.signature_version))
        self.connectionCls.host = host

        super().__init__(key, secret, secure, host, port, api_version, region, **kwargs)

    def _ex_connection_class_kwargs(self):
        kwargs = {}
        kwargs["signature_version"] = self.signature_version
        return kwargs



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from base64 import b64encode

from libcloud.common.base import Connection, JsonResponse
from libcloud.container.base import ContainerImage

__all__ = ["RegistryClient", "HubClient"]


class DockerHubConnection(Connection):
    responseCls = JsonResponse

    def __init__(
        self,
        host,
        username=None,
        password=None,
        secure=True,
        port=None,
        url=None,
        timeout=None,
        proxy_url=None,
        backoff=None,
        retry_delay=None,
    ):
        super().__init__(
            secure=secure,
            host=host,
            port=port,
            url=url,
            timeout=timeout,
            proxy_url=proxy_url,
            backoff=backoff,
            retry_delay=retry_delay,
        )
        self.username = username
        self.password = password

    def add_default_headers(self, headers):
        headers["Content-Type"] = "application/json"
        if self.username is not None:
            authstr = "Basic " + str(
                b64encode(("{}:{}".format(self.username, self.password)).encode("latin1")).strip()
            )
            headers["Authorization"] = authstr
        return headers


class RegistryClient:
    """
    A client for the Docker v2 registry API
    """

    connectionCls = DockerHubConnection

    def __init__(self, host, username=None, password=None, **kwargs):
        """
        Construct a Docker registry client

        :param host: Your registry endpoint, e.g. 'registry.hub.docker.com'
        :type  host: ``str``

        :param username: (optional) Your registry account username
        :type  username: ``str``

        :param password: (optional) Your registry account password
        :type  password: ``str``
        """
        self.connection = self.connectionCls(host, username, password, **kwargs)

    def list_images(self, repository_name, namespace="library", max_count=100):
        """
        List the tags (versions) in a repository

        :param  repository_name: The name of the repository e.g. 'ubuntu'
        :type   repository_name: ``str``

        :param  namespace: (optional) The docker namespace
        :type   namespace: ``str``

        :param  max_count: The maximum number of records to return
        :type   max_count: ``int``

        :return: A list of images
        :rtype: ``list`` of :class:`libcloud.container.base.ContainerImage`
        """
        path = "/v2/repositories/{}/{}/tags/?page=1&page_size={}".format(
            namespace,
            repository_name,
            max_count,
        )
        response = self.connection.request(path)
        images = []
        for image in response.object["results"]:
            images.append(self._to_image(repository_name, image))
        return images

    def get_repository(self, repository_name, namespace="library"):
        """
        Get the information about a specific repository

        :param  repository_name: The name of the repository e.g. 'ubuntu'
        :type   repository_name: ``str``

        :param  namespace: (optional) The docker namespace
        :type   namespace: ``str``

        :return: The details of the repository
        :rtype: ``object``
        """
        path = "/v2/repositories/{}/{}/".format(namespace, repository_name)
        response = self.connection.request(path)
        return response.object

    def get_image(self, repository_name, tag="latest", namespace="library"):
        """
        Get an image from a repository with a specific tag

        :param repository_name: The name of the repository, e.g. ubuntu
        :type  repository_name: ``str``

        :param  tag: (optional) The image tag (defaults to latest)
        :type   tag: ``str``

        :param  namespace: (optional) The docker namespace
        :type   namespace: ``str``

        :return: A container image
        :rtype: :class:`libcloud.container.base.ContainerImage`
        """
        path = "/v2/repositories/{}/{}/tags/{}/".format(namespace, repository_name, tag)
        response = self.connection.request(path)
        return self._to_image(repository_name, response.object)

    def _to_image(self, repository_name, obj):
        path = "{}/{}:{}".format(self.connection.host, repository_name, obj["name"])
        return ContainerImage(
            id=obj["id"],
            path=path,
            name=path,
            version=obj["name"],
            extra={"full_size": obj["full_size"]},
            driver=None,
        )


class HubClient(RegistryClient):
    """
    A client for the Docker Hub API

    The hub is based on the v2 registry API
    """

    host = "registry.hub.docker.com"

    def __init__(self, username=None, password=None, **kwargs):
        """
        Construct a Docker hub client

        :param username: (optional) Your Hub account username
        :type  username: ``str``

        :param password: (optional) Your hub account password
        :type  password: ``str``
        """
        super().__init__(self.host, username, password, **kwargs)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import ModuleType
from typing import TYPE_CHECKING, Type, Union

from libcloud.storage.types import OLD_CONSTANT_TO_NEW_MAPPING, Provider
from libcloud.common.providers import get_driver as _get_provider_driver
from libcloud.common.providers import set_driver as _set_provider_driver

if TYPE_CHECKING:
    from libcloud.storage.base import StorageDriver

DRIVERS = {
    Provider.DUMMY: ("libcloud.storage.drivers.dummy", "DummyStorageDriver"),
    Provider.CLOUDFILES: (
        "libcloud.storage.drivers.cloudfiles",
        "CloudFilesStorageDriver",
    ),
    Provider.OPENSTACK_SWIFT: (
        "libcloud.storage.drivers.cloudfiles",
        "OpenStackSwiftStorageDriver",
    ),
    Provider.S3: ("libcloud.storage.drivers.s3", "S3StorageDriver"),
    Provider.S3_US_EAST2: ("libcloud.storage.drivers.s3", "S3USEast2StorageDriver"),
    Provider.S3_US_WEST: ("libcloud.storage.drivers.s3", "S3USWestStorageDriver"),
    Provider.S3_US_WEST_OREGON: (
        "libcloud.storage.drivers.s3",
        "S3USWestOregonStorageDriver",
    ),
    Provider.S3_US_GOV_WEST: (
        "libcloud.storage.drivers.s3",
        "S3USGovWestStorageDriver",
    ),
    Provider.S3_CN_NORTH: ("libcloud.storage.drivers.s3", "S3CNNorthStorageDriver"),
    Provider.S3_CN_NORTHWEST: (
        "libcloud.storage.drivers.s3",
        "S3CNNorthWestStorageDriver",
    ),
    Provider.S3_EU_WEST: ("libcloud.storage.drivers.s3", "S3EUWestStorageDriver"),
    Provider.S3_EU_WEST2: ("libcloud.storage.drivers.s3", "S3EUWest2StorageDriver"),
    Provider.S3_EU_CENTRAL: ("libcloud.storage.drivers.s3", "S3EUCentralStorageDriver"),
    Provider.S3_EU_NORTH1: ("libcloud.storage.drivers.s3", "S3EUNorth1StorageDriver"),
    Provider.S3_AP_SOUTH: ("libcloud.storage.drivers.s3", "S3APSouthStorageDriver"),
    Provider.S3_AP_SOUTHEAST: ("libcloud.storage.drivers.s3", "S3APSEStorageDriver"),
    Provider.S3_AP_SOUTHEAST2: ("libcloud.storage.drivers.s3", "S3APSE2StorageDriver"),
    Provider.S3_AP_NORTHEAST: ("libcloud.storage.drivers.s3", "S3APNE1StorageDriver"),
    Provider.S3_AP_NORTHEAST1: ("libcloud.storage.drivers.s3", "S3APNE1StorageDriver"),
    Provider.S3_AP_NORTHEAST2: ("libcloud.storage.drivers.s3", "S3APNE2StorageDriver"),
    Provider.S3_SA_EAST: ("libcloud.storage.drivers.s3", "S3SAEastStorageDriver"),
    Provider.S3_CA_CENTRAL: ("libcloud.storage.drivers.s3", "S3CACentralStorageDriver"),
    Provider.S3_RGW: ("libcloud.storage.drivers.rgw", "S3RGWStorageDriver"),
    Provider.S3_RGW_OUTSCALE: (
        "libcloud.storage.drivers.rgw",
        "S3RGWOutscaleStorageDriver",
    ),
    Provider.NINEFOLD: ("libcloud.storage.drivers.ninefold", "NinefoldStorageDriver"),
    Provider.GOOGLE_STORAGE: (
        "libcloud.storage.drivers.google_storage",
        "GoogleStorageDriver",
    ),
    Provider.NIMBUS: ("libcloud.storage.drivers.nimbus", "NimbusStorageDriver"),
    Provider.LOCAL: ("libcloud.storage.drivers.local", "LocalStorageDriver"),
    Provider.AZURE_BLOBS: (
        "libcloud.storage.drivers.azure_blobs",
        "AzureBlobsStorageDriver",
    ),
    Provider.KTUCLOUD: ("libcloud.storage.drivers.ktucloud", "KTUCloudStorageDriver"),
    Provider.AURORAOBJECTS: (
        "libcloud.storage.drivers.auroraobjects",
        "AuroraObjectsStorageDriver",
    ),
    Provider.BACKBLAZE_B2: (
        "libcloud.storage.drivers.backblaze_b2",
        "BackblazeB2StorageDriver",
    ),
    Provider.ALIYUN_OSS: ("libcloud.storage.drivers.oss", "OSSStorageDriver"),
    Provider.DIGITALOCEAN_SPACES: (
        "libcloud.storage.drivers.digitalocean_spaces",
        "DigitalOceanSpacesStorageDriver",
    ),
    Provider.MINIO: ("libcloud.storage.drivers.minio", "MinIOStorageDriver"),
    Provider.SCALEWAY: ("libcloud.storage.drivers.scaleway", "ScalewayStorageDriver"),
    Provider.OVH: ("libcloud.storage.drivers.ovh", "OvhStorageDriver"),
}


def get_driver(provider):
    # type: (Union[Provider, str]) -> Type[StorageDriver]
    deprecated_constants = OLD_CONSTANT_TO_NEW_MAPPING
    return _get_provider_driver(
        drivers=DRIVERS, provider=provider, deprecated_constants=deprecated_constants
    )


def set_driver(provider, module, klass):
    # type: (Union[Provider, str], ModuleType, type) -> Type[StorageDriver]
    return _set_provider_driver(drivers=DRIVERS, provider=provider, module=module, klass=klass)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import random
import string
import unittest

from libcloud.utils.py3 import httplib
from libcloud.common.gandi import GandiException
from libcloud.test.secrets import GANDI_PARAMS
from libcloud.compute.types import NodeState
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.compute.drivers.gandi import GandiNodeDriver
from libcloud.test.common.test_gandi import BaseGandiMockHttp


class GandiTests(unittest.TestCase):
    node_name = "test2"

    def setUp(self):
        GandiNodeDriver.connectionCls.conn_class = GandiMockHttp
        GandiMockHttp.type = None
        self.driver = GandiNodeDriver(*GANDI_PARAMS)

    def test_list_nodes(self):
        nodes = self.driver.list_nodes()
        self.assertTrue(len(nodes) > 0)
        self.assertTrue(len(nodes[0].public_ips) > 1)

    def test_list_locations(self):
        loc = list(filter(lambda x: "france" in x.country.lower(), self.driver.list_locations()))[0]
        self.assertEqual(loc.country, "France")

    def test_list_images(self):
        loc = list(filter(lambda x: "france" in x.country.lower(), self.driver.list_locations()))[0]
        images = self.driver.list_images(loc)
        self.assertTrue(len(images) > 2)

    def test_list_sizes(self):
        sizes = self.driver.list_sizes()
        self.assertTrue(len(sizes) >= 1)

    def test_destroy_node_running(self):
        nodes = self.driver.list_nodes()
        test_node = list(filter(lambda x: x.state == NodeState.RUNNING, nodes))[0]
        self.assertTrue(self.driver.destroy_node(test_node))

    def test_destroy_node_halted(self):
        nodes = self.driver.list_nodes()
        test_node = list(filter(lambda x: x.state == NodeState.TERMINATED, nodes))[0]
        self.assertTrue(self.driver.destroy_node(test_node))

    def test_reboot_node(self):
        nodes = self.driver.list_nodes()
        test_node = list(filter(lambda x: x.state == NodeState.RUNNING, nodes))[0]
        self.assertTrue(self.driver.reboot_node(test_node))

    def test_create_node(self):
        login = "libcloud"
        passwd = "".join(random.choice(string.ascii_letters) for i in range(10))

        # Get france datacenter
        loc = list(filter(lambda x: "france" in x.country.lower(), self.driver.list_locations()))[0]

        # Get a debian image
        images = self.driver.list_images(loc)
        images = [x for x in images if x.name.lower().startswith("debian")]
        img = list(filter(lambda x: "5" in x.name, images))[0]

        # Get a configuration size
        size = self.driver.list_sizes()[0]
        node = self.driver.create_node(
            name=self.node_name,
            login=login,
            password=passwd,
            image=img,
            location=loc,
            size=size,
        )
        self.assertEqual(node.name, self.node_name)

    def test_create_volume(self):
        loc = list(filter(lambda x: "france" in x.country.lower(), self.driver.list_locations()))[0]
        volume = self.driver.create_volume(size=1024, name="libcloud", location=loc)
        self.assertEqual(volume.name, "libcloud")
        self.assertEqual(volume.size, 1024)

    def test_list_volumes(self):
        disks = self.driver.list_volumes()
        self.assertTrue(len(disks) > 0)

    def test_destroy_volume(self):
        volumes = self.driver.list_volumes()
        test_vol = list(filter(lambda x: x.name == "test_disk", volumes))[0]
        self.assertTrue(self.driver.destroy_volume(test_vol))

    def test_attach_volume(self):
        disks = self.driver.list_volumes()
        nodes = self.driver.list_nodes()
        res = self.driver.attach_volume(nodes[0], disks[0])
        self.assertTrue(res)

    def test_detach_volume(self):
        disks = self.driver.list_volumes()
        nodes = self.driver.list_nodes()
        res = self.driver.detach_volume(nodes[0], disks[0])
        self.assertTrue(res)

    def test_ex_list_interfaces(self):
        ifaces = self.driver.ex_list_interfaces()
        self.assertTrue(len(ifaces) > 0)

    def test_ex_attach_interface(self):
        ifaces = self.driver.ex_list_interfaces()
        nodes = self.driver.list_nodes()
        res = self.driver.ex_node_attach_interface(nodes[0], ifaces[0])
        self.assertTrue(res)

    def test_ex_detach_interface(self):
        ifaces = self.driver.ex_list_interfaces()
        nodes = self.driver.list_nodes()
        res = self.driver.ex_node_detach_interface(nodes[0], ifaces[0])
        self.assertTrue(res)

    def test_ex_snapshot_disk(self):
        disks = self.driver.list_volumes()
        self.assertTrue(self.driver.ex_snapshot_disk(disks[2]))
        self.assertRaises(GandiException, self.driver.ex_snapshot_disk, disks[0])

    def test_ex_update_disk(self):
        disks = self.driver.list_volumes()
        self.assertTrue(self.driver.ex_update_disk(disks[0], new_size=4096))

    def test_list_key_pairs(self):
        keys = self.driver.list_key_pairs()
        self.assertTrue(len(keys) > 0)

    def test_get_key_pair(self):
        key = self.driver.get_key_pair(10)
        self.assertEqual(key.name, "testkey")

    def test_import_key_pair_from_string(self):
        key = self.driver.import_key_pair_from_string("testkey", "12345")
        self.assertEqual(key.name, "testkey")
        self.assertEqual(key.extra["id"], 10)

    def test_delete_key_pair(self):
        response = self.driver.delete_key_pair(10)
        self.assertTrue(response)

    def test_ex_get_node(self):
        node = self.driver.ex_get_node(34951)
        self.assertEqual(node.name, "test2")

    def test_ex_get_volume(self):
        volume = self.driver.ex_get_volume(1263)
        self.assertEqual(volume.name, "libcloud")


class GandiRatingTests(unittest.TestCase):

    """Tests where rating model is involved"""

    node_name = "test2"

    def setUp(self):
        GandiNodeDriver.connectionCls.conn_class = GandiMockRatingHttp
        GandiMockRatingHttp.type = None
        self.driver = GandiNodeDriver(*GANDI_PARAMS)

    def test_list_sizes(self):
        sizes = self.driver.list_sizes()
        self.assertEqual(len(sizes), 4)

    def test_create_node(self):
        login = "libcloud"
        passwd = "".join(random.choice(string.ascii_letters) for i in range(10))

        # Get france datacenter
        loc = list(filter(lambda x: "france" in x.country.lower(), self.driver.list_locations()))[0]

        # Get a debian image
        images = self.driver.list_images(loc)
        images = [x for x in images if x.name.lower().startswith("debian")]
        img = list(filter(lambda x: "5" in x.name, images))[0]

        # Get a configuration size
        size = self.driver.list_sizes()[0]
        node = self.driver.create_node(
            name=self.node_name,
            login=login,
            password=passwd,
            image=img,
            location=loc,
            size=size,
        )
        self.assertEqual(node.name, self.node_name)


class GandiMockHttp(BaseGandiMockHttp):
    fixtures = ComputeFileFixtures("gandi")

    def _xmlrpc__hosting_datacenter_list(self, method, url, body, headers):
        body = self.fixtures.load("datacenter_list.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_image_list(self, method, url, body, headers):
        body = self.fixtures.load("image_list_dc0.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_vm_list(self, method, url, body, headers):
        body = self.fixtures.load("vm_list.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_ip_list(self, method, url, body, headers):
        body = self.fixtures.load("ip_list.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_account_info(self, method, url, body, headers):
        body = self.fixtures.load("account_info.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_vm_info(self, method, url, body, headers):
        body = self.fixtures.load("vm_info.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_vm_delete(self, method, url, body, headers):
        body = self.fixtures.load("vm_delete.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__operation_info(self, method, url, body, headers):
        body = self.fixtures.load("operation_info.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_vm_create_from(self, method, url, body, headers):
        body = self.fixtures.load("vm_create_from.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_vm_reboot(self, method, url, body, headers):
        body = self.fixtures.load("vm_reboot.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_vm_stop(self, method, url, body, headers):
        body = self.fixtures.load("vm_stop.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_iface_list(self, method, url, body, headers):
        body = self.fixtures.load("iface_list.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_disk_list(self, method, url, body, headers):
        body = self.fixtures.load("disk_list.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_vm_iface_attach(self, method, url, body, headers):
        body = self.fixtures.load("iface_attach.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_vm_iface_detach(self, method, url, body, headers):
        body = self.fixtures.load("iface_detach.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_vm_disk_attach(self, method, url, body, headers):
        body = self.fixtures.load("disk_attach.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_vm_disk_detach(self, method, url, body, headers):
        body = self.fixtures.load("disk_detach.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_disk_create(self, method, url, body, headers):
        body = self.fixtures.load("disk_create.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_disk_create_from(self, method, url, body, headers):
        body = self.fixtures.load("disk_create_from.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_disk_info(self, method, url, body, headers):
        body = self.fixtures.load("disk_info.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_disk_update(self, method, url, body, headers):
        body = self.fixtures.load("disk_update.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_disk_delete(self, method, url, body, headers):
        body = self.fixtures.load("disk_delete.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_ssh_info(self, method, url, body, headers):
        body = self.fixtures.load("ssh_info.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_ssh_list(self, method, url, body, headers):
        body = self.fixtures.load("ssh_list.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_ssh_create(self, method, url, body, headers):
        body = self.fixtures.load("ssh_info.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_ssh_delete(self, method, url, body, headers):
        body = self.fixtures.load("ssh_delete.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


class GandiMockRatingHttp(BaseGandiMockHttp):

    """Fixtures needed for tests related to rating model"""

    fixtures = ComputeFileFixtures("gandi")

    def _xmlrpc__hosting_datacenter_list(self, method, url, body, headers):
        body = self.fixtures.load("datacenter_list.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_image_list(self, method, url, body, headers):
        body = self.fixtures.load("image_list_dc0.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_vm_create_from(self, method, url, body, headers):
        body = self.fixtures.load("vm_create_from.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__operation_info(self, method, url, body, headers):
        body = self.fixtures.load("operation_info.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc__hosting_vm_info(self, method, url, body, headers):
        body = self.fixtures.load("vm_info.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    # Specific to rating tests
    def _xmlrpc__hosting_account_info(self, method, url, body, headers):
        body = self.fixtures.load("account_info_rating.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

from libcloud.dns.providers import Provider as DnsProvider
from libcloud.dns.providers import get_driver as get_dns_driver
from libcloud.backup.providers import Provider as BackupProvider
from libcloud.backup.providers import get_driver as get_backup_driver
from libcloud.compute.providers import Provider as ComputeProvider
from libcloud.compute.providers import get_driver as get_compute_driver
from libcloud.storage.providers import Provider as StorageProvider
from libcloud.storage.providers import get_driver as get_storage_driver
from libcloud.container.providers import Provider as ContainerProvider
from libcloud.container.providers import get_driver as get_container_driver
from libcloud.loadbalancer.providers import Provider as LoadBalancerProvider
from libcloud.loadbalancer.providers import get_driver as get_loadbalancer_driver


class DriverType:
    """Backup-as-a-service driver"""

    BACKUP = BackupProvider

    """ Compute-as-a-Service driver """
    COMPUTE = ComputeProvider

    """ Container-as-a-Service driver """
    CONTAINER = ContainerProvider

    """ DNS service provider driver """
    DNS = DnsProvider

    """ Load balancer provider-driver """
    LOADBALANCER = LoadBalancerProvider

    """ Storage-as-a-Service driver """
    STORAGE = StorageProvider


DriverTypeFactoryMap = {
    DriverType.BACKUP: get_backup_driver,
    DriverType.COMPUTE: get_compute_driver,
    DriverType.CONTAINER: get_container_driver,
    DriverType.DNS: get_dns_driver,
    DriverType.LOADBALANCER: get_loadbalancer_driver,
    DriverType.STORAGE: get_storage_driver,
}  # type: Dict


class DriverTypeNotFoundError(KeyError):
    def __init__(self, type):
        self.message = "Driver type '%s' not found." % type

    def __repr__(self):
        return self.message


def get_driver(type, provider):
    """
    Get a driver
    """
    try:
        return DriverTypeFactoryMap[type](provider)
    except KeyError:
        raise DriverTypeNotFoundError(type)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.common.base import XmlResponse, ConnectionKey

__all__ = ["ZonomiException", "ZonomiResponse", "ZonomiConnection"]

# Endpoint for Zonomi API.
API_HOST = "zonomi.com"

SPECIAL_ERRORS = [
    "Not found.",
    "ERROR: This zone is already in your zone list.",
    "Record not deleted.",
]


class ZonomiException(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message
        self.args = (code, message)

    def __str__(self):
        return "{} {}".format(self.code, self.message)

    def __repr__(self):
        return "ZonomiException {} {}".format(self.code, self.message)


class ZonomiResponse(XmlResponse):
    errors = None
    objects = None

    def __init__(self, response, connection):
        self.errors = []
        super().__init__(response=response, connection=connection)
        self.objects, self.errors = self.parse_body_and_errors()
        if self.errors:
            raise self._make_excp(self.errors[0])

    def parse_body_and_errors(self):
        error_dict = {}
        actions = None
        result_counts = None
        action_childrens = None
        data = []
        errors = []
        xml_body = super().parse_body()

        # pylint: disable=no-member
        # Error handling
        if xml_body.text is not None and xml_body.tag == "error":
            error_dict["ERRORCODE"] = self.status
            if xml_body.text.startswith("ERROR: No zone found for"):
                error_dict["ERRORCODE"] = "404"
                error_dict["ERRORMESSAGE"] = "Not found."
            else:
                error_dict["ERRORMESSAGE"] = xml_body.text
            errors.append(error_dict)

        # Data handling
        children = list(xml_body)
        if len(children) == 3:
            result_counts = children[1]
            actions = children[2]

        if actions is not None:
            actions_childrens = list(actions)
            action = actions_childrens[0]
            action_childrens = list(action)

        if action_childrens is not None:
            for child in action_childrens:
                if child.tag == "zone" or child.tag == "record":
                    data.append(child.attrib)

        if result_counts is not None and result_counts.attrib.get("deleted") == "1":
            data.append("DELETED")

        if (
            result_counts is not None
            and result_counts.attrib.get("deleted") == "0"
            and action.get("action") == "DELETE"
        ):
            error_dict["ERRORCODE"] = self.status
            error_dict["ERRORMESSAGE"] = "Record not deleted."
            errors.append(error_dict)

        return (data, errors)

    def success(self):
        return len(self.errors) == 0

    def _make_excp(self, error):
        """
        :param error: contains error code and error message
        :type error: dict
        """
        return ZonomiException(error["ERRORCODE"], error["ERRORMESSAGE"])


class ZonomiConnection(ConnectionKey):
    host = API_HOST
    responseCls = ZonomiResponse

    def add_default_params(self, params):
        """
        Adds default parameters to perform a request,
        such as api_key.
        """
        params["api_key"] = self.key

        return params

    def add_default_headers(self, headers):
        """
        Adds default headers needed to perform a successful
        request such as Content-Type, User-Agent.
        """
        headers["Content-Type"] = "text/xml;charset=UTF-8"

        return headers
